# -*- coding: utf-8 -*-

"""
This file contains all the tunables available in mycelyso.

You can set them via :code:`-t Name=value` on the command line.
"""

from tunable import Tunable
from .processing import binarization as binarization_module


class NodeEndpointMergeRadius(Tunable):
    """ Radius in which endpoints are going to be merged [µm] """
    default = 0.5


class NodeJunctionMergeRadius(Tunable):
    """ Radius in which junctions are going to be merged [µm] """
    default = 0.5


class NodeLookupRadius(Tunable):
    """ Radius in which nodes will be searched for found pixel structures [µm] """
    default = 0.5


class NodeLookupCutoffRadius(Tunable):
    """ Radius at which nodes will be ignored if they are further away [µm] """
    default = 2.5


class NodeTrackingJunctionShiftRadius(Tunable):
    """ Maximum search radius for junctions [µm·h⁻¹] """
    default = 5.0


class NodeTrackingEndpointShiftRadius(Tunable):
    """ Maximum search radius for endpoints [µm·h⁻¹] """
    default = 100.0


class CropWidth(Tunable):
    """ Crop value (horizontal) of the image [pixels] """
    default = 0


class CropHeight(Tunable):
    """ Crop value (vertical) of the image [pixels] """
    default = 0


class BoxDetection(Tunable):
    """ Whether to run the rectangular microfluidic growth structure detection as ROI detection """
    default = False


class StoreImage(Tunable):
    """ Whether to store images in the resulting HDF5. This leads to a potentially much larger output file. """
    default = False


class SkipBinarization(Tunable):
    """ Whether to directly use the input image as binary mask. Use in case external binarization is desired. """
    default = False


class CleanUpGaussianSigma(Tunable):
    """Clean up step: Sigma [µm] used for Gaussian filter"""
    default = 0.075


class CleanUpGaussianThreshold(Tunable):
    """Clean up step: Threshold used after Gaussian filter (values range from 0 to 1)"""
    default = 0.5


class CleanUpHoleFillSize(Tunable):
    """Clean up step: Maximum size of holes [µm²] which will be filled"""
    default = 1.0


class RemoveSmallStructuresSize(Tunable):
    """Remove structures up to this size [µm²]"""
    default = 10.0


class BorderArtifactRemovalBorderSize(Tunable):
    """Remove structures, whose centroid lies within that distance [µm] of a border"""
    default = 10.0


class TrackingMaximumRelativeShrinkage(Tunable):
    """ Tracking, maximal relative shrinkage """
    default = 0.2


class TrackingMinimumTipElongationRate(Tunable):
    """ Tracking, minimum tip elongation rate [µm·h⁻¹]"""
    default = -0.0


class TrackingMaximumTipElongationRate(Tunable):
    """ Tracking, maximum tip elongation rate [µm·h⁻¹] """
    default = 100.0


class TrackingMaximumCoverage(Tunable):
    """ Tracking, maximum covered area ratio at which tracking is still performed """
    default = 0.2


class TrackingMinimumTrackedPointCount(Tunable):
    """ Tracking, minimal time steps in track filter [#] """
    default = 5


class TrackingMinimalMaximumLength(Tunable):
    """ Tracking, minimal hyphae end length in track filter [µm] """
    default = 10.0


class TrackingMinimalGrownLength(Tunable):
    """ Tracking, minimal hyphae gained length in track filter [µm] """
    default = 5.0


class ThresholdingTechnique(Tunable):
    """ Binarization method to use, for available methods see documentation (mycelyso.processing.binarization) """

    @classmethod
    def test(cls, value):
        return value in dir(binarization_module)

    default = "experimental_thresholding"


class ThresholdingParameters(Tunable):
    """ Parameters for the used binarization method, passed as key1:value1,key2:value2,... string """
    default = ""



# -*- coding: utf-8 -*-
"""
The maintenance_interrupt submodule provides the install_maintenance_interrupt function, which allows under *NIX systems
to interrupt the running process and spawn a IPython shell within the running process.
"""

try:
    from signal import signal, SIGUSR2
except ImportError:
    signal, SIGUSR2 = None, None

import traceback


def maintenance_interrupt(the_signal, frame):

    print("Interrupted at:")
    print(''.join(traceback.format_stack(frame)))
    print("Have a look at frame.f_globals and frame.f_locals")
    try:
        from IPython import embed
        embed()
    except ImportError:
        embed = None
        from code import interact
        interact(local=locals())
    print("... continuing")


def install_maintenance_interrupt():
    if signal is None or SIGUSR2 is None:
        return  # no functionality in windows
    signal(SIGUSR2, maintenance_interrupt)




# -*- coding: utf-8 -*-
"""
The imagestack module contains the base functionality for nD bio image opening with metadata support.
"""

from .imagestack import ImageStack, Dimensions, FloatFilter, MinMaxFilter, UnwrapFilter




# -*- coding: utf-8 -*-
"""
The OME-TIFF reader module implements an ImageStack able to open OME-TIFF images, by using the tifffile module
to open the particular tiff file in general and custom XML processing code to parse the OME annotation in particular.
"""

from ..imagestack import Dimensions
from .tiff import TiffImageStack

from xml.etree import cElementTree as ElementTree


class OMETiffImageStack(TiffImageStack):
    extensions = ('.ome.tif', '.ome.tiff',)

    priority = 500

    def open(self, location, **kwargs):
        super(OMETiffImageStack, self).open(location, **kwargs)

        self.fp = self.tiff.pages[0]
        if not self.fp.is_ome:
            raise RuntimeError('Not an OMETiffStack')
        self.xml = None
        self.ns = ''
        self.xml_str = self.fp.tags['image_description'].value

        try:
            self.treat_z_as_mp = bool(self.parameters['treat_z_as_mp'])
        except KeyError:
            self.treat_z_as_mp = False

        self.images = self._parse_ome_xml(self.xml_str)

        an_image = next(iter(next(iter(self.images.values()))))

        self.set_dimensions_and_sizes(
            [Dimensions.Time, Dimensions.PositionXY, Dimensions.PositionZ, Dimensions.Channel],
            [an_image['SizeT'], len(self.images), an_image['SizeZ'], an_image['SizeC']])

    @staticmethod
    def pixel_attrib_sanity_check(pa):
        """

        :param pa:
        :raise RuntimeError:
        """
        if pa['BigEndian'] == 'true':
            raise RuntimeError("Unsupported Pixel format")
        if pa['Interleaved'] == 'true':
            raise RuntimeError("Unsupported Pixel format")

    def _parse_ome_xml(self, xml):
        try:  # bioformats seem to copy some (wrongly encoded) strings verbatim
            root = ElementTree.fromstring(xml)
        except ElementTree.ParseError:
            root = ElementTree.fromstring(xml.decode('iso-8859-1').encode('utf-8'))

        self.xml = root

        self.ns = ns = root.tag.split('}')[0][1:]

        # sometimes string properties creep up, but then we don't care as we don't plan on using them ...
        def float_or_int(s):
            """

            :param s:
            :return:
            """
            try:
                if '.' in s:
                    return float(s)
                else:
                    return int(s)
            except ValueError:
                return s

        keep_pa = {'SizeZ', 'SizeY', 'SizeX', 'SignificantBits', 'PhysicalSizeX', 'PhysicalSizeY', 'SizeC', 'SizeT'}

        images = {}

        if self.treat_z_as_mp:  # handling for mal-encoded files
            image_nodes = [n for n in root.getchildren() if n.tag == ElementTree.QName(ns, 'Image')]
            # there will be only one image node
            imn = image_nodes[0]

            pixels = [n for n in imn.getchildren() if n.tag == ElementTree.QName(ns, 'Pixels')][0]

            pa = pixels.attrib

            self.pixel_attrib_sanity_check(pa)

            pai = list({k: v for k, v in pa.items() if k in keep_pa}.items())

            tiff_data = {
                (n.attrib['FirstC'], n.attrib['FirstT'], n.attrib['FirstZ']): n.attrib
                for n in pixels.getchildren() if n.tag == ElementTree.QName(ns, 'TiffData')
            }
            planes = [dict(
                list(n.attrib.items()) +
                list(tiff_data[(n.attrib['TheC'], n.attrib['TheT'], n.attrib['TheZ'])].items()) + pai
            ) for n in pixels.getchildren() if n.tag == ElementTree.QName(ns, 'Plane')]

            planes = [{k: float_or_int(v) for k, v in p.items()} for p in planes]
            multipoints = range(planes[0]['SizeZ'])
            images = {mp: [p for p in planes if p['TheZ'] == mp] for mp in multipoints}
            # more fixing

            def _correct_attributes(inner_p, inner_planes):
                inner_p['PositionX'] = inner_planes[0]['PositionX']
                inner_p['PositionY'] = inner_planes[0]['PositionY']
                inner_p['PositionZ'] = inner_planes[0]['PositionZ']
                inner_p['TheZ'] = 0
                return inner_p

            images = {mp: [_correct_attributes(p, planes) for p in planes] for mp, planes in images.items()}

        else:
            image_nodes = [n for n in root.getchildren() if n.tag == ElementTree.QName(ns, 'Image')]
            for n, imn in enumerate(image_nodes):
                pixels = [n for n in imn.getchildren() if n.tag == ElementTree.QName(ns, 'Pixels')][0]

                pa = pixels.attrib

                self.pixel_attrib_sanity_check(pa)

                pai = list({k: v for k, v in pa.items() if k in keep_pa}.items())

                tiff_data = {
                    (n.attrib['FirstC'], n.attrib['FirstT'], n.attrib['FirstZ']): n.attrib
                    for n in pixels.getchildren() if n.tag == ElementTree.QName(ns, 'TiffData')
                }
                planes = [dict(
                    list(n.attrib.items()) +
                    list(tiff_data[(n.attrib['TheC'], n.attrib['TheT'], n.attrib['TheZ'])].items()) + pai
                ) for n in pixels.getchildren() if n.tag == ElementTree.QName(ns, 'Plane')]

                planes = [{k: float_or_int(v) for k, v in p.items()} for p in planes]

                images[n] = planes

        return images

    def get_data(self, what):

        try:
            channel = what[Dimensions.Channel]
        except KeyError:
            channel = 0

        tps = self.images[what[Dimensions.PositionXY]]
        tp = [tp for tp in tps if tp['TheT'] == what[Dimensions.Time] and tp['TheC'] == channel][0]

        return self.tiff.pages[tp['IFD']].asarray()

    def get_meta(self, what):

        tps = self.images[what[Dimensions.PositionXY]]
        image = [tp for tp in tps if tp['TheT'] == what[Dimensions.Time]][0]

        position = self.__class__.Position(x=image['PositionX'], y=image['PositionY'], z=image['PositionZ'],)
        meta = self.__class__.Metadata(time=image['DeltaT'], position=position, calibration=image['PhysicalSizeX'])
        return meta



# coding=utf-8

import numpy as np
import os
from os import getpid
from collections import namedtuple
from itertools import product

# noinspection PyCompatibility
from urllib.parse import urlparse, parse_qs


class Dimensions(object):

    @classmethod
    def all(cls):
        return cls.Dimension.__subclasses__()

    @classmethod
    def all_by_char(cls):
        return {d.char: d for d in cls.all()}

    @classmethod
    def by_char(cls, char):
        return cls.all_by_char()[char]

    # noinspection PyClassHasNoInit
    class Dimension(object):
        char = 'd'

    # noinspection PyClassHasNoInit
    class Time(Dimension):
        char = 't'

    # noinspection PyClassHasNoInit
    class PositionXY(Dimension):
        char = 'r'  # region

    Position = PositionXY

    # noinspection PyClassHasNoInit
    class PositionZ(Dimension):
        char = 'z'

    # noinspection PyClassHasNoInit
    class Channel(Dimension):
        char = 'c'

    # noinspection PyClassHasNoInit
    class Width(Dimension):
        char = 'w'

    # noinspection PyClassHasNoInit
    class Height(Dimension):
        char = 'h'


class UnsupportedImageError(NotImplementedError):
    pass


class ImageStackAPI(object):
    def get_data(self, what):
        pass

    def get_meta(self, what):
        return {}

    def init(self):
        pass

    def open(self, location, **kwargs):
        pass

    def notify_fork(self):
        pass


class ImageStackFilter(object):
    def filter(self, image):
        return image


class Image(np.ndarray):
    def __new__(cls, input_array, meta=None):
        obj = np.asarray(input_array).view(cls)
        obj.meta = meta
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

        self.meta = getattr(obj, 'meta', None)

    def __array_wrap__(self, out_arr, context=None):
        # noinspection PyArgumentList
        return np.ndarray.__array_wrap__(self, out_arr, context)


class FloatFilter(ImageStackFilter):
    def filter(self, image):
        return image.astype(np.float32, copy=False)


class MinMaxFilter(ImageStackFilter):
    def filter(self, image):
        float_image = image.astype(np.float32, copy=True)
        float_image -= float_image.min()
        float_image /= float_image.max()
        return float_image


class UnwrapFilter(ImageStackFilter):
    def filter(self, image):
        return image.view(np.ndarray)


Metadata = namedtuple('Metadata', ['time', 'position', 'calibration'])
Position = namedtuple('Position', ['x', 'y', 'z'])


class ImageStack(ImageStackAPI):

    Metadata = Metadata
    Position = Position

    priority = 0
    schemes = ('file',)
    extensions = ('',)

    class ImageStackAccessBridge(object):
        def __init__(self, parent, purpose):
            self.parent = parent
            self.purpose = purpose

        def __getitem__(self, item):
            return self.parent.bridged_access(self.purpose, item)

    # noinspection PyProtectedMember
    class ImageStackView(object):
        @property
        def order(self):
            if self._order:
                return self._order
            else:
                return self.parent._all_dimensions[:]

        @order.setter
        def order(self, order):
            self._order = order

        def __init__(self, parent, pinned_indices=None, order=None, filters=None):
            self.parent = parent

            if pinned_indices:
                self.pinned_indices = pinned_indices
            else:
                self.pinned_indices = {}

            self.next_pinned_indices = None

            self._order = order

            self.meta = ImageStack.ImageStackAccessBridge(self, 'meta')
            self.data = ImageStack.ImageStackAccessBridge(self, 'data')

            if filters:
                self.filters = filters
            else:
                self.filters = []

        def copy(self):
            return self.__class__(self.parent, self.pinned_indices.copy(), self.order, self.filters)

        def filter(self, *what):
            self_copy = self.copy()
            if isinstance(what[0], tuple) or isinstance(what[0], list):
                what = what[0]

            for w in what:
                self_copy.filters.append(w)

            return self_copy

        def pin_indices(self, item):
            self_copy = self.copy()

            self_copy.begin()

            if isinstance(item, tuple):
                for n, i in enumerate(item):
                    self_copy.pin_index(n, i)
            elif isinstance(item, slice):
                self_copy.pin_index(0, item)
            elif isinstance(item, int):
                self_copy.pin_index(0, item)
            else:
                raise IndexError('...')

            self_copy.commit()

            return self_copy

        def __getitem__(self, item):
            self_copy = self.pin_indices(item)

            if self_copy.is_fixed():
                filters = []
                for a_filter in self.filters:
                    if issubclass(a_filter, ImageStackFilter):
                        filter_instance = a_filter()
                        filters.append(filter_instance.filter)
                    else:
                        filters.append(a_filter)

                return self.parent.perform_get_data(self_copy.collect(), filters)
            else:
                return self_copy

        def collect(self):
            return {
                k: v * self.parent.get_subsampling(k)
                for k, v in self.pinned_indices.items() if k in self.parent._all_dimensions
            }

        def is_fixed(self):
            for item in self.pinned_indices.values():
                if type(item) != int:
                    return False
            return True

        def pin_index(self, index, to_what):
            n = 0
            item = None

            for item in self.order:
                try:
                    v = self.pinned_indices[item]
                except KeyError:
                    v = self.next_pinned_indices[item] = self.pinned_indices[item] = False

                if v is False or isinstance(v, range):
                    if index == n:
                        break
                    n += 1

            try:
                size = self.parent.size[item]
            except KeyError:
                size = 1

            to_what_raw = to_what

            if isinstance(to_what, slice):
                start, stop, step = to_what.start, to_what.stop, to_what.step
                if start is None:
                    start = 0
                if stop is None:
                    stop = size
                if step is None:
                    step = 1

                to_what = range(start, stop, step)

            if self.next_pinned_indices[item] is False:
                self.next_pinned_indices[item] = to_what
            else:
                self.next_pinned_indices[item] = self.next_pinned_indices[item][to_what_raw]

        def commit(self):
            self.pinned_indices = self.next_pinned_indices

        def begin(self):
            self.next_pinned_indices = self.pinned_indices.copy()

        def bridged_access(self, entry_point, what):
            if entry_point == 'meta':
                self_copy = self.pin_indices(what)

                if self_copy.is_fixed():
                    return self.parent.perform_get_meta(self_copy.collect())
                else:
                    return self_copy

            elif entry_point == 'data':
                return self.__getitem__(what)

        def view(self, *new_order):
            if isinstance(new_order[0], list):
                new_order = new_order[0]

            self_copy = self.copy()
            self_copy.order = list(new_order)
            return self_copy

        @property
        def dimensions(self):
            return [
                item
                for item in self.order
                if item not in self.pinned_indices or type(self.pinned_indices[item]) != int
            ]
            # return self.order

        @property
        def size(self):
            result = {}
            for k in self.order:
                try:
                    result[k] = self.parent._sizes[self.parent._all_dimensions.index(k)]
                except ValueError:
                    result[k] = 1
            return result

        @property
        def sizes(self):
            sizes = self.size
            return [sizes[k] for k in self.dimensions]

        def __len__(self):
            result = 1
            for size in self.sizes:
                result *= size
            return result

        @property
        def every_index(self):
            for tup in product(*(range(s) for s in self.sizes)):
                yield tup

        @property
        def every_dict(self):
            dim = self.dimensions
            for tup in self.every_index:
                yield dict(zip(dim, tup))

        @property
        def every_index_image(self):
            for tup in self.every_index:
                yield tup, self.__getitem__(tup)

        @property
        def every_dict_image(self):
            dim = self.dimensions
            for tup in self.every_index:
                yield dict(zip(dim, tup)), self.__getitem__(tup)

    def __new__(cls, *args, **kwargs):
        if cls != ImageStack:
            return object.__new__(cls)

        what = args[0]
        to_open = urlparse(what)

        if os.name == 'nt':
            # windows peculiarity
            drive_prefix = "%s:" % (to_open.scheme,)
            if len(to_open.scheme) == 1 and os.path.isdir(drive_prefix):
                # noinspection PyProtectedMember
                to_open = to_open._replace(scheme='')._replace(path=drive_prefix + to_open.path)

        if to_open.scheme == '':
            # noinspection PyProtectedMember
            to_open = to_open._replace(scheme='file')

        def recursive_subclasses(class_, collector):
            collector.add(class_)
            for inner_class_ in class_.__subclasses__():
                recursive_subclasses(inner_class_, collector)
            return collector

        subclasses = recursive_subclasses(cls, set()) - {ImageStack}

        hits = list(sorted((
            class_ for class_ in subclasses
            if
            sum(to_open.scheme == scheme for scheme in class_.schemes) > 0 and
            sum(to_open.path.endswith(ext) for ext in class_.extensions) > 0),
            key=lambda c: c.priority))

        parameters = {k: (v if len(v) > 1 else v[0]) for k, v in parse_qs(to_open.query).items()}

        for hit in hits:
            try:
                result = hit()
                result.uri = what
                result.parameters = parameters
                result.parse_parameters_before()
                result.open(to_open, **parameters)
                result.parse_parameters_after()
                return result
            except UnsupportedImageError:
                pass

        raise RuntimeError('No suitable image plugin found.')

    __init___called = False

    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        if self.__init___called:
            return

        self.parent = self

        self.pid = getpid()

        self.__init___called = True
        self._sizes = []
        self._all_dimensions = []

        self.parameters = {}
        self.subsampling = {}

        self.root_view = self.__class__.ImageStackView(self)

        self.init()

    def parse_parameters_before(self):
        self.subsampling = {}
        dimensions = Dimensions.all_by_char()
        for param, value in self.parameters.items():
            if param.startswith('subsample_'):
                d = param.replace('subsample_', '')
                if d in dimensions:
                    self.subsampling[dimensions[d]] = int(value)

    def parse_parameters_after(self):
        dimensions = Dimensions.all_by_char()
        for param, value in self.parameters.items():
            if param in dimensions:
                dim = dimensions[param]
                self.root_view.pinned_indices[dim] = int(value)
                self.root_view = self.root_view.view([_d for _d in self.root_view.order if _d != dim])

    def get_subsampling(self, dim):
        try:
            return self.subsampling[dim]
        except KeyError:
            return 1

    def __getitem__(self, item):
        return self.root_view.__getitem__(item)

    def __getattr__(self, item):
        if hasattr(self.root_view, item):
            return getattr(self.root_view, item)

    def __len__(self):
        return self.root_view.__len__()

    def set_dimensions_and_sizes(self, dimensions, sizes):
        self._real_dimensions = dimensions
        self._real_sizes = sizes

        self._all_dimensions = dimensions
        self._sizes = [s // self.get_subsampling(dim) for dim, s in zip(dimensions, sizes)]

    def perform_get_data(self, what, filters=None):
        self.check_fork()
        data = self.get_data(what)

        xy_subsample = [self.parent.get_subsampling(Dimensions.Width), self.parent.get_subsampling(Dimensions.Height)]
        if xy_subsample[0] != xy_subsample[1]:
            raise RuntimeError('Asymmetric subsampling for width/height currently unsupported!')

        xy_subsample = xy_subsample[0]
        if xy_subsample != 1:
            data = data[::xy_subsample, ::xy_subsample]

        data = data.view(Image)

        data.meta = self.perform_get_meta(what)

        if filters:
            for a_callable_filter in filters:
                data = a_callable_filter(data)

        return data

    def perform_get_meta(self, what):
        self.check_fork()
        meta = self.get_meta(what)

        if isinstance(meta, Metadata):
            # time_offset allows to shift all times by an offset,
            # eg. when multiple files belong to one experiment, but all start at 0 albeit consecutive
            time_offset_str = 'time_offset'
            if time_offset_str in self.parameters:
                meta = meta._replace(time=meta.time + float(self.parameters[time_offset_str]))

        return meta

    def check_fork(self):
        if self.pid != getpid():
            self.notify_fork()
        self.pid = getpid()




# -*- coding: utf-8 -*-
"""
The pixelgraphs module contains various functions to work with skeleton images, and treating the paths of the
skeleton as graphs, which can be walked along.
"""

import numpy as np

from scipy import ndimage as ndi


_COUNTING_KERNEL = np.array(
    [[1, 1, 1],
     [1, 0, 1],
     [1, 1, 1]], dtype=np.uint8
)

_COUNTING_KERNEL.flags.writeable = False

N, NW, W, SW, S, SE, E, NE = 2, 1, 128, 64, 32, 16, 8, 4

# noinspection PyPep8
_NEIGHBORHOOD_KERNEL = np.array(
    [[NW,  N, NE],
     [ W,  0,  E],
     [SW,  S, SE]], dtype=np.uint8
)

# a constant
_NEIGHBORHOOD_KERNEL.flags.writeable = False

X = 1
Y = 0

# remember: Y, X
# noinspection PyPep8
_NEIGHBORS = {
    #    YY  XX
    N:  [-1,  0],
    NW: [-1, -1],
    W:  [ 0, -1],
    SW: [ 1, -1],
    S:  [ 1,  0],
    SE: [ 1,  1],
    E:  [ 0,  1],
    NE: [-1,  1]
}

_INVERSE_NEIGHBORS = {
    N: S,
    NW: SE,
    W: E,
    SW: NE,
    S: N,
    SE: NW,
    E: W,
    NE: SW,
}


def get_connectivity_map(binary):
    """
    Returns a 'connectivity map', where each value represents the count of neighbors at a position.
    
    :param binary: Binary input image 
    :return: 
    """
    return binary * ndi.correlate(binary.astype(np.uint8), _COUNTING_KERNEL, mode='constant')


def get_neighborhood_map(binary):
    """
    Returns a 'neighborhood map', where each value binary encodes the connections at a point.
    
    :param binary: Binary input image 
    :return: 
    """
    return binary * ndi.correlate(binary.astype(np.uint8), _NEIGHBORHOOD_KERNEL, mode='constant')


def get_next_neighbor(num):
    """
    Returns the coordinates represented by a numeric neighbor bit. 
    
    :param num: Neighbor bit 
    :return: Shift (r, c)
    """
    return _NEIGHBORS[num]


def get_all_neighbor_nums(num):
    """
    Return all set neighbor bits in num.
    
    :param num: Neighborhood representation. 
    :return: Array of values
    """
    return [value for value, shift in _NEIGHBORS.items() if (num & value) == value]


def get_all_neighbors(num):
    """
    Return positions for all set neighbor bits in num
    
    :param num: Neighborhood representation
    :return: Array of shifts
    """
    return [shift for value, shift in _NEIGHBORS.items() if (num & value) == value]


def get_inverse_neighbor_shift(num):
    """
    Get the shift corresponding to the inverse direction represented by num. 
    
    :param num: Neighborhood bit
    :return: Shift (r, c)
    """
    return _INVERSE_NEIGHBORS[num]


def is_edge(connectivity):
    """
    Returns True if connectivity corresponds an edge (is two).
    
    :param connectivity: Scalar or matrix 
    :return: Boolean or matrix of boolean
    """
    return connectivity == 2


def is_junction(connectivity):
    """
    Returns True if connectivity corresponds a junction (is greater than two).
    
    :param connectivity: Scalar or matrix 
    :return: Boolean or matrix of boolean
    """
    return connectivity > 2


def is_end(connectivity):
    """
    Returns True if connectivity corresponds to an endpoint (is one).
    
    :param connectivity: Scalar or matrix 
    :return: Boolean or matrix of boolean 
    """
    return connectivity == 1


def where2d(image):
    """
    numpy.where for 2D matrices.
    
    :param image: Input images
    :return: Coordinate list where image is non-zero
    
    >>> where2d(np.array([[ 0, 0, 0],
    ...                   [ 0, 1, 1],
    ...                   [ 0, 0, 0]]))
    array([[1, 1],
           [1, 2]])
    """

    hits = np.where(image.ravel())[0]

    result = np.c_[hits, hits]

    result[:, 0] //= image.shape[1]
    result[:, 1] %= image.shape[1]

    return result



import numpy as np
from itertools import tee
from scipy.ndimage import uniform_filter1d
from scipy.spatial.ckdtree import cKDTree as KDTree


def calculate_length(points, times=1, w=5):
    """
    Calculates the length of a path.
    
    Paths sampled from pixel grids may contain notable measuring error, if euclidean distances are calculated
    naively. This method uses an adapted approach from [Cornelisse1984]_, by repeatedly smoothing
    the coordinates with a moving average filter before calculating the euclidean distance.
    
    .. [Cornelisse1984] Cornelisse and van den Berg (1984) Journal of Microscopy
       `10.1111/j.1365-2818.1984.tb00544.x <https://dx.doi.org/10.1111/j.1365-2818.1984.tb00544.x>`_
    
    :param points: Input points, a numpy array (X, 2) 
    :param times: Times smoothing should be applied
    :param w: window width of the moving average filter
    :return: Length of the input path
    
    >>> calculate_length(np.array([[1.0, 1.0],
    ...                            [5.0, 5.0]]))
    5.6568542494923806
    """
    # adapted method from Cornelisse and van den Berg
    if (len(points) - 2) > w:
        for _ in range(times):
            points[:, 0] = uniform_filter1d(points[:, 0], w, mode='nearest')
            points[:, 1] = uniform_filter1d(points[:, 1], w, mode='nearest')

    result = np.sqrt((np.diff(points, axis=0) ** 2.0).sum(axis=1)).sum()

    return result


def clean_by_radius(points, radius=15.0):
    """
    Bins points by radius and returns only one per radius, removing duplicates.
    
    :param points: Input points 
    :param radius: Radius
    :return: Filtered points
    
    >>> clean_by_radius(np.array([[1.0, 1.0],
    ...                           [1.1, 1.1],
    ...                           [9.0, 9.0]]), radius=1.5)
    array([[ 1.,  1.],
           [ 9.,  9.]])
    """
    if len(points) == 0:
        return points
    tree = KDTree(points)
    mapping = tree.query_ball_tree(tree, radius)
    unique_indices = np.unique(list(l[0] for l in sorted(mapping)))
    return points[unique_indices]


# from the itertools help https://docs.python.org/2/library/itertools.html
def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)
# end



# -*- coding: utf-8 -*-
"""
The steps file contains various generally reusable image processing steps.
"""

import hashlib
from base64 import b64encode

import numpy as np
import scipy.ndimage as ndi
from ..imagestack.imagestack import Dimensions
from ..application.application import Meta

from mfisp_boxdetection import find_box
from molyso.generic.rotation import find_rotation, rotate_image
from molyso.generic.registration import translation_2x1d, shift_image
from ..pipeline.executor import Skip, Collected
from ..misc.h5writer import CompressedObject


class Delete(object):
    pass


class Compress(object):
    pass


def set_result(**kwargs):
    def _inner(result):
        for k, v in kwargs.items():
            if v is Delete:
                del result[k]
            elif v is Compress:
                result[k] = CompressedObject(result[k])
            else:
                result[k] = v
        return result

    return _inner


# noinspection PyUnusedLocal
def image_source(ims, meta, image=None):
    return ims.view(Dimensions.Position, Dimensions.Time)[meta.pos, meta.t]


# noinspection PyUnusedLocal
def calculate_image_sha256_hash(image, image_sha256_hash=None):
    hasher = hashlib.sha256()
    hasher.update(image.tobytes())
    hash_value = b64encode(hasher.digest()).decode()
    return hash_value


def image_to_ndarray(image):
    return np.array(image)


# noinspection PyUnusedLocal
def pull_metadata_from_image(image, timepoint=None, position=None, calibration=None):
    return image, image.meta.time, image.meta.position, image.meta.calibration


_substract_start_frame_start_images = {}


# noinspection PyUnusedLocal
def substract_start_frame(meta, ims, reference_timepoint, image, subtracted_image=None):
    gaussian_blur_radius = 15.0

    if meta.pos not in _substract_start_frame_start_images:
        reference = image_source(ims, Meta(t=reference_timepoint, pos=meta.pos))
        blurred = ndi.gaussian_filter(reference, gaussian_blur_radius)
        _substract_start_frame_start_images[meta.pos] = blurred
    else:
        blurred = _substract_start_frame_start_images[meta.pos]

    image = image.astype(np.float32)

    image /= blurred

    image -= image.min()
    image /= image.max()

    return image


_box_cache_fft_cache = {}
_box_cache_boxes = {}
_box_cache_angles = {}


def _box_detection_get_parameters(ims, timepoint, pos):
    reference = image_source(ims, Meta(t=timepoint, pos=pos))
    angle = find_rotation(reference)
    reference = rotate_image(reference, angle)

    shift, fft_a = translation_2x1d(image_a=reference, image_b=reference, return_a=True)

    try:

        reference = reference.astype(np.float32)

        cleaned_reference = reference - ndi.uniform_filter(reference, 25)  # "box" blur
        cleaned_reference[cleaned_reference < 0] = 0
        cleaned_reference /= cleaned_reference.max()

        # qimshow(cleaned_reference)

        crop = find_box(cleaned_reference, throw=True, subsample=2)
        t, b, l, r = crop
        cleaned_reference[t, :] = 1
        cleaned_reference[b, :] = 1
        cleaned_reference[:, l] = 1
        cleaned_reference[:, r] = 1

    except RuntimeError:
        # noinspection PyCompatibility
        raise Skip(Meta(pos=pos, t=Collected)) from None

    return angle, fft_a, crop


# noinspection PyUnusedLocal
def box_detection(ims, image, meta, reference_timepoint, shift=None, crop=None, angle=None):
        # probably implement a voting scheme?

        if meta.pos not in _box_cache_boxes:
            _box_cache_angles[meta.pos], _box_cache_fft_cache[meta.pos], _box_cache_boxes[meta.pos] = \
                _box_detection_get_parameters(ims, reference_timepoint, meta.pos)

        angle = _box_cache_angles[meta.pos]
        image = rotate_image(image, angle)

        shift, = translation_2x1d(image_a=None, image_b=image, ffts_a=_box_cache_fft_cache[meta.pos])
        crop = _box_cache_boxes[meta.pos]

        return shift, crop, angle


def create_boxcrop_from_subtracted_image(subtracted_image, shift, angle, crop, result):
    result.shift_x = shift[0]
    result.shift_y = shift[1]

    result.crop_t, result.crop_b, result.crop_l, result.crop_r = crop
    t, b, l, r = crop

    subtracted_image = rotate_image(subtracted_image, angle)
    subtracted_image = shift_image(subtracted_image, shift, background='blank')

    box = subtracted_image[t:b, l:r]

    result.image = box

    return result


def rescale_image_to_uint8(image):
    image = image.astype(np.float32)
    image -= image.min()
    image /= image.max()

    image *= 255.0

    return image.astype(np.uint8)



# -*- coding: utf-8 -*-
"""
The processpool submodule contains a custom process-pool, with some additional features not present in standard
Python processpool management, e.g. timeouts.
"""

from time import sleep
import datetime
import traceback
import heapq

from multiprocessing import Process, Pipe, cpu_count

from multiprocessing.pool import Pool

exception_debugging = False


class Future(object):
    command = None
    args = None
    kwargs = None

    value = None
    error = None

    process = None
    pool = None

    priority = 0

    status = None

    timeout = 0
    started_at = None

    def __init__(self):
        pass

    def __lt__(self, other):
        # dummy
        return self.priority < other.priority

    def wait(self, time=None):
        pass

    def fail(self):
        self.status, (self.value, self.error) = True, (None, None)

        if self.process:
            self.process.terminate()
            self.pool.report_broken_process(self.process)
        else:
            self.pool.future_became_ready(self)

    def ready(self):
        if self.status:
            return self.status

        if self.process is None:
            # not scheduled yet
            return False

        if self.timeout > 0:
            now = datetime.datetime.now()
            if (now - self.started_at).total_seconds() > self.timeout:
                # we reached a hard timeout

                try:
                    self.process.terminate()
                    # kill(self.process.pid, SIGINT)
                    sleep(0.25)
                except Exception as e:
                    print(repr(e))

                # self.process.terminate()

                self.pool.report_broken_process(self.process)
                self.status, (self.value, self.error) = \
                    True, (None, RuntimeError('Process took longer than specified timeout and was terminated.'))
                return

        if not self.process.is_alive():
            exitcode = self.process.exitcode
            self.pool.report_broken_process(self.process)
            self.status, (self.value, self.error) = \
                True, (None, RuntimeError('Process trying to work on this future died. Exitcode: %d' % (exitcode,)))
            return

        self.status, (self.value, self.error) = self.process.ready()

        if self.status:
            self.pool.future_became_ready(self)

        return self.status

    def get(self):
        not_ready = True
        while not_ready:
            not_ready = not self.ready()

        if self.error:
            raise self.error
        else:
            return self.value

    def dispatch(self):
        self.started_at = datetime.datetime.now()
        self.process.dispatch()


class WrappedException(RuntimeError):
    exception = None
    message = None

    def __init__(self, exception, message=''):
        self.exception = exception
        self.message = message


class FutureProcess(Process):
    STARTUP = 0
    RUN = 1
    STOP = 2

    def __init__(self):
        super(FutureProcess, self).__init__()
        self.future = None
        self.pipe_parent_end, self.pipe_child_end = Pipe()

    def run(self):
        while True:
            command_type, command, args, kwargs = self.pipe_child_end.recv()

            if command_type == FutureProcess.STOP:
                break

            if command_type == FutureProcess.STARTUP and command is None:
                continue

            result = None
            exc = None

            try:
                result = command(*args, **kwargs)
            except Exception as e:
                exc = WrappedException(e, traceback.format_exc())

            if command_type == FutureProcess.STARTUP:
                continue

            self.pipe_child_end.send((result, exc,))

    def send_command(self, *args):
        self.pipe_parent_end.send(args)

    def dispatch(self):
        self.send_command(FutureProcess.RUN, self.future.command, self.future.args, self.future.kwargs)

    def ready(self):
        if self.pipe_parent_end.poll():
            return True, self.pipe_parent_end.recv()
        else:
            return False, (None, None,)


class SimpleProcessPool(object):
    def new_process(self):
        p = FutureProcess()
        p.start()
        p.send_command(*self.startup_message)
        return p

    def fill_pool(self):
        for _ in range(self.count - (len(self.waiting_processes) + len(self.active_processes))):
            self.waiting_processes.add(self.new_process())

    def __init__(self, processes=0, initializer=None, initargs=None, initkwargs=None, future_timeout=0):

        if initargs is None:
            initargs = ()

        if initkwargs is None:
            initkwargs = {}

        self.startup_message = (FutureProcess.STARTUP, initializer, initargs, initkwargs)

        if processes == 0:
            processes = cpu_count()

        self.future_timeout = future_timeout
        self.count = processes

        self.waiting_processes = set()
        self.active_processes = set()

        self.fill_pool()

        self.active_futures = set()
        self.waiting_futures = []  # priority queue

        self.closing = False

    def close(self):
        self.closing = True

        self.schedule()

    def report_broken_process(self, p):
        f = p.future
        if p in self.active_processes:
            self.active_processes.remove(p)

        if p in self.waiting_processes:
            # print("found a process where it does not belong", p)
            self.waiting_processes.remove(p)

        if f in self.active_futures:
            self.active_futures.remove(f)

        if f in self.waiting_futures:
            # remove one entry, rebuild
            self.waiting_futures.remove(f)
            heapq.heapify(self.waiting_futures)

        # if f in self.waiting_futures:
        #    #print("found a future where it does not belong", f)
        #    self.waiting_futures.remove(f)

        self.fill_pool()

        self.schedule()

    def apply(self, command, *args, **kwargs):
        return self.advanced_apply(command=command, args=args, kwargs=kwargs)

    def advanced_apply(self, command=None, priority=0, args=None, kwargs=None):
        if args is None:
            args = tuple()
        if kwargs is None:
            kwargs = {}

        f = Future()
        f.command = command
        f.args = args
        f.kwargs = kwargs

        f.priority = priority

        f.timeout = self.future_timeout

        f.pool = self

        # self.waiting_futures.add(f)
        heapq.heappush(self.waiting_futures, f)

        self.schedule()

        return f

    # ugly signature
    def apply_async(self, fun, args=None, kwargs=None):
        if args is None:
            args = {}
        if kwargs is None:
            kwargs = {}
        return self.apply(fun, *args, **kwargs)

    def future_became_ready(self, f):
        if f in self.active_futures:
            self.active_futures.remove(f)

        if f.process in self.active_processes:
            self.active_processes.remove(f.process)

        if f.process:
            self.waiting_processes.add(f.process)

        self.schedule()

    def schedule(self):
        for f in self.active_futures.copy():
            f.ready()

        while len(self.waiting_processes) > 0:
            if len(self.waiting_futures) == 0:
                break

            # f = self.waiting_futures.pop()
            f = heapq.heappop(self.waiting_futures)

            p = self.waiting_processes.pop()

            f.process = p
            p.future = f

            self.active_processes.add(p)
            self.active_futures.add(f)

            f.dispatch()

        if self.closing:
            while len(self.waiting_processes) > 0:
                p = self.waiting_processes.pop()
                p.send_command(FutureProcess.STOP, None, [], {})
                self.active_processes.add(p)


class DuckTypedApplyResult(object):
    def __init__(self, callable_):
        self.value = None
        self.called = False
        self.callable = callable_

    # noinspection PyMethodMayBeStatic
    def ready(self):
        return True

    def wait(self, timeout):
        pass

    def fail(self):
        self.value = None
        self.called = True

    def get(self):
        if not self.called:
            self.value = None
            try:
                self.value = self.callable()
            except Exception as e:
                if exception_debugging:
                    raise

                raise WrappedException(e, traceback.format_exc())

        return self.value


# noinspection PyAbstractClass,PyUnusedLocal
class NormalPool(Pool):
    def advanced_apply(self, command, args, **kwargs):
        return self.apply(func=command, args=args)


# noinspection PyUnusedLocal
class InProcessFakePool(object):
    @staticmethod
    def advanced_apply(command, args, **kwargs):
        def _bind_it(inner_args):
            def _perform():
                return command(*inner_args)

            return _perform

        return DuckTypedApplyResult(_bind_it(args))

    def close(self):
        pass



# -*- coding: utf-8 -*-
"""
The application submodule contains the base class for applications, as well auxiliary functions.
"""

import argparse
import logging
import multiprocessing
# noinspection PyUnresolvedReferences
import sys
import numpy as np

from tunable import TunableManager
from ..misc.hacks.maintenance_interrupt import install_maintenance_interrupt
from ..misc.hacks.recursionlimit_raise import *
from ..misc.hacks.multiprocessing_patch import *

from collections import namedtuple

from ..imagestack.imagestack import ImageStack, Dimensions
# noinspection PyUnresolvedReferences
from ..imagestack.readers import *
from ..pipeline import PipelineExecutor


def parse_range(s, maximum=0):
    """
    Parses a range string.
    
    :param s:
    :param maximum:
    :return:
    """
    maximum -= 1
    splits = s.replace(' ', '').replace(';', ',').split(',')

    ranges = []
    remove = []

    not_values = False

    for frag in splits:
        if frag[0] == '~':
            not_values = not not_values
            frag = frag[1:]

        if '-' in frag:
            f, t = frag.split('-')

            interval = 1

            if '%' in t:
                t, _interval = t.split('%')
                interval = int(_interval)

            if t == '':
                t = maximum

            f, t = int(f), int(t)

            t = min(t, maximum)

            parsed_fragment = range(f, t + 1, interval)
        else:
            parsed_fragment = [int(frag)]

        if not_values:
            remove += parsed_fragment
        else:
            ranges += parsed_fragment

    return list(sorted(set(ranges) - set(remove)))


def prettify_numpy_array(arr, space_or_prefix):
    """
    Returns a properly indented string representation of a numpy array.

    :param arr:
    :param space_or_prefix:
    :return:
    """
    six_spaces = ' ' * 6
    prepared = repr(np.array(arr)).replace(')', '').replace('array(', six_spaces)
    if isinstance(space_or_prefix, int):
        return prepared.replace(six_spaces, ' ' * space_or_prefix)
    else:
        return space_or_prefix + prepared.replace(six_spaces, ' ' * len(space_or_prefix)).lstrip()


class AppInterface(object):
    """
    Interface to be implemented by Apps utilizing pilyso's App infrastructure.
    """
    def options(self):
        return {}

    def arguments(self, argparser):
        return

    def setup(self, pipeline_executor):
        return


Meta = namedtuple('Meta', ['pos', 't'])


class App(AppInterface):
    """
    Base class implementing most of the App "in the back" works.
    """
    @staticmethod
    def _internal_option_defaults():
        return {
            'name':
                "processor",
            'description':
                "processor",
            'banner':
                "",
            'pipeline': None
        }

    _internal_options = None

    @property
    def internal_options(self):
        if self._internal_options is None:
            self._internal_options = self._internal_option_defaults()
            self._internal_options.update(self.options())
        return self._internal_options

    _log = None

    @property
    def log(self):
        if self._log is None:
            self._log = logging.getLogger(self.internal_options['name'])
        return self._log

    def _create_argparser(self):
        argparser = argparse.ArgumentParser(description=self.internal_options['description'])

        def _error(message=''):
            argparser.print_help()
            self.log.error("command line argument error: %s", message)
            sys.exit(1)

        argparser.error = _error

        return argparser

    @staticmethod
    def _arguments(argparser):
        argparser.add_argument('input', metavar='input', type=str, help="input file")
        argparser.add_argument('-m', '--module', dest='modules', type=str, default=None, action='append')
        argparser.add_argument('-n', '--processes', dest='processes', default=-1, type=int)
        argparser.add_argument('--prompt', '--prompt', dest='wait_on_start', default=False, action='store_true')
        argparser.add_argument('-tp', '--timepoints', dest='timepoints', default='0-', type=str)
        argparser.add_argument('-mp', '--positions', dest='positions', default='0-', type=str)

        TunableManager.register_argparser(argparser)

    def handle_args(self):
        pass

    args = None

    def main(self):
        """
        Main entry point for App apps.
        :return: 
        """
        logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(name)s %(levelname)s %(message)s")

        install_maintenance_interrupt()

        argparser = self._create_argparser()

        self._arguments(argparser)
        self.arguments(argparser)

        self.log.info(self.internal_options['banner'])
        self.args = argparser.parse_args()

        self.handle_args()

        if self.args.wait_on_start:
            _ = input("Press enter to continue.")

        self.log.info("Started %s.", self.internal_options['name'])

        ims = ImageStack(self.args.input).view(Dimensions.PositionXY, Dimensions.Time)

        # self._setup_modules()

        self.positions = parse_range(self.args.positions, maximum=ims.size[Dimensions.PositionXY])
        self.timepoints = parse_range(self.args.timepoints, maximum=ims.size[Dimensions.Time])

        self.log.info("Tunable Hash: %s" % (TunableManager.get_hash()))

        self.log.info(
            "Beginning Processing:\n%s\n%s",
            prettify_numpy_array(self.positions,  "Positions : "),
            prettify_numpy_array(self.timepoints, "Timepoints: ")
        )

        if self.args.processes < 0:
            self.args.processes = multiprocessing.cpu_count()
        elif self.args.processes == 0:
            self.args.processes = False

        self.pe = PipelineExecutor()

        self.pe.multiprocessing = self.args.processes

        self.pe.set_workload(Meta, Meta(pos=self.positions, t=self.timepoints), Meta(t=1, pos=2))

        self.setup(self.pe)

        self.run()

        self.pe.close()

        self.log.info("Finished %s.", self.internal_options['name'])

    def run(self):
        self.pe.run()

    # default implementation
    def setup(self, pipeline_executor):
        pipeline_executor.initialize(self.internal_options['pipeline'], self.args)




# -*- coding: utf-8 -*-
"""
The multiprocessing_patch module monkey-patches the send/receive functions of Python's multiprocessing module,
so they work properly if blocks larger than 2**31 have to be transferred.
"""

# thank you python, in times of exascale computing someone thought 2**31 bytes oughta be enough ...

import multiprocessing.connection
import struct
import io

SEND_MAX = 2**31 - 8*1024 - 1


# noinspection PyProtectedMember
def _new_send(self, buf, write=multiprocessing.connection.Connection._write):
    remaining = len(buf)
    while True:
        try:
            n = write(self._handle, buf)
        except InterruptedError:
            continue
        remaining -= n
        if remaining == 0:
            break
        buf = buf[n:]


# noinspection PyProtectedMember
def _new_recv(self, size, read=multiprocessing.connection.Connection._read):
    buf = io.BytesIO()
    handle = self._handle
    remaining = size
    while remaining > 0:
        try:
            if remaining > SEND_MAX:
                chunk = read(handle, SEND_MAX)
            else:
                chunk = read(handle, remaining)
        except InterruptedError:
            continue
        n = len(chunk)
        if n == 0:
            if remaining == size:
                raise EOFError
            else:
                raise OSError("got end of file during message")
        buf.write(chunk)
        remaining -= n
    return buf


def _new_send_bytes(self, buf):
    n = len(buf)
    # For wire compatibility with 3.2 and lower
    header = struct.pack("!Q", n)
    if n > 16380:
        # The payload is large so Nagle's algorithm won't be triggered
        # and we'd better avoid the cost of concatenation.
        chunks = [header, buf]
    elif n > 0:
        # Issue # 20540: concatenate before sending, to avoid delays due
        # to Nagle's algorithm on a TCP socket.
        chunks = [header + buf]
    else:
        # This code path is necessary to avoid "broken pipe" errors
        # when sending a 0-length buffer if the other end closed the pipe.
        chunks = [header]
    for chunk in chunks:
        self._send(chunk)


def _new_recv_bytes(self, maxsize=None):
    buf = self._recv(8)
    size, = struct.unpack("!Q", buf.getvalue())
    if maxsize is not None and size > maxsize:
        return None
    return self._recv(size)


multiprocessing.connection.Connection._send_bytes = _new_send_bytes
multiprocessing.connection.Connection._recv_bytes = _new_recv_bytes
multiprocessing.connection.Connection._send = _new_send
multiprocessing.connection.Connection._recv = _new_recv



"""
The CZI reader module implements an ImageStack able to open CZI (Carl Zeiss Imaging) image stacks,
by using the czifile module.
"""

from __future__ import division, unicode_literals, print_function

from ..imagestack import ImageStack, Dimensions

from xml.etree import cElementTree as ElementTree

import warnings

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    # code tend to throw warnings because of missing C extensions
    from .external.czifile import CziFile, TimeStamps, etree as czifile_etree


def _is_dimension_oi(d):
    return d in {'T', 'C', 'Z', 'S'}


def _dim_shape_to_dict(dim, shape):
    if not isinstance(dim, str):
        dim = dim.decode()
    return dict(zip(list(dim), shape))


def _get_subblock_position(subblock):
    return _dim_shape_to_dict(subblock.axes, subblock.start)


def _get_subblock_identifier(subblock):
    return _normalize(_get_subblock_position(subblock))


def _get_image_from_subblock(subblock):
    return subblock.data_segment().data().reshape([s for s in subblock.stored_shape if s != 1])


def _normalize(d):
    return tuple([(k, v,) for k, v in sorted(d.items()) if _is_dimension_oi(k)])


def _only_existing_dim(reference, test):
    test = test.copy()
    for k in list(test.keys()):
        if k not in reference:
            del test[k]
    return test


class CziImageStack(ImageStack):
    extensions = ('.czi',)

    priority = 500

    def open(self, location, **kwargs):

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.czi = CziFile(location.path)

        self.frames = {
            _get_subblock_identifier(subblock): subblock for subblock in self.czi.filtered_subblock_directory
            }

        self.size = _dim_shape_to_dict(self.czi.axes, self.czi.shape)

        self.metadata = ElementTree.fromstring(czifile_etree.tostring(self.czi.metadata))

        # /ImageDocument
        calibration_x = float(
            self.metadata.find("./Metadata/Scaling/Items/Distance[@Id='X']/Value").text
        ) * 1E6

        calibration_y = float(
            self.metadata.find("./Metadata/Scaling/Items/Distance[@Id='Y']/Value").text
        ) * 1E6

        assert calibration_x == calibration_y

        self.calibration = calibration_x

        timestamps = None

        for entry in self.czi.attachment_directory:
            entry_data = entry.data_segment().data()
            if isinstance(entry_data, TimeStamps):
                timestamps = entry_data
                break

        self.timestamps = timestamps

        positions = []

        for scene in sorted(
                self.metadata.findall("./Metadata/Information/Image/Dimensions/S/Scenes/Scene"),
                key=lambda scene: int(scene.attrib['Index'])):
            center_position = next(child.text for child in scene.getchildren() if child.tag == "CenterPosition")
            x, y = center_position.split(',')
            positions.append((float(x), float(y)))

        self.positions = positions

        self.set_dimensions_and_sizes(
            [Dimensions.Time, Dimensions.PositionXY, Dimensions.PositionZ, Dimensions.Channel],
            [self.size.get('T', 1), self.size.get('S', 1), 1, self.size.get('C', 1)]
        )

    # noinspection PyProtectedMember
    def notify_fork(self):
        self.czi._fh.close()
        self.czi._fh.open()

    def get_data(self, what):
        channel = what.get(Dimensions.Channel, 0)
        position = what.get(Dimensions.PositionXY, 0)
        time = what.get(Dimensions.Time, 0)

        return _get_image_from_subblock(self.frames[_normalize(
            _only_existing_dim(self.size, dict(C=channel, S=position, T=time))
        )])

    def get_meta(self, what):
        try:
            time = float(self.timestamps[what[Dimensions.Time]])
        except TypeError:
            time = what[Dimensions.Time]

        try:
            position = (
                self.positions[what[Dimensions.PositionXY]][0],
                self.positions[what[Dimensions.PositionXY]][1],
                0.0,
            )
        except KeyError:
            position = (float('nan'), float('nan'), 0.0)

        meta = self.__class__.Metadata(
            time=time,
            position=position,
            calibration=self.calibration)

        return meta




# -*- coding: utf-8 -*-
"""
The steps module contains most of the individual, albeit mycelyso-specific processing steps.
"""


import warnings
from itertools import product, chain

import numpy as np
from scipy.stats import linregress

from ..tunables import CleanUpGaussianSigma, CleanUpGaussianThreshold, CleanUpHoleFillSize, \
    RemoveSmallStructuresSize, BorderArtifactRemovalBorderSize, TrackingMaximumRelativeShrinkage, \
    TrackingMinimumTipElongationRate, TrackingMaximumTipElongationRate, TrackingMaximumCoverage, \
    TrackingMinimumTrackedPointCount, TrackingMinimalMaximumLength, TrackingMinimalGrownLength, \
    ThresholdingTechnique, ThresholdingParameters

from skimage.morphology import remove_small_holes, remove_small_objects, skeletonize as sk_skeletonize
from skimage.measure import label, regionprops

from scipy import ndimage as ndi

import networkx as nx

from ..misc.graphml import to_graphml_string
from ..misc.util import pairwise
from ..misc.regression import prepare_optimized_regression
from ..pilyso.application import Collected, Meta, Skip
from ..pilyso.pipeline.pipeline import get_argnames_and_defaults

from ..processing import binarization as binarization_module

from .pixelframe import PixelFrame
from .nodeframe import NodeFrame

try:
    import matplotlib.pyplot as pyplot
except ImportError:
    pyplot = None


def qimshow(image, cmap='gray'):
    """
    Debug function, quickly shows the passed image via matplotlibs imshow-facilities.
    
    :param image: 
    :param cmap: 
    :return: 
    """
    if not pyplot:
        raise RuntimeError('matplotlib not installed.')
    fig = pyplot.figure()
    ax = fig.add_subplot(111)
    range_ = ax.imshow(image, cmap=cmap, interpolation='none')

    def _format_coords(x, y):
        try:
            y, x = int(y + 0.5), int(x + 0.5)
            if y < 0 or x < 0:
                raise IndexError
            value = image[y, x]
        except IndexError:
            value = float('nan')
        return 'x=%d y=%d value=%1.4f' % (x, y, value,)

    ax.format_coord = _format_coords
    pyplot.colorbar(range_)
    pyplot.show()


# noinspection PyUnusedLocal
def set_empty_crops(image, crop_t=None, crop_b=None, crop_l=None, crop_r=None):
    """
    Defines crop parameters based upon image size, effectively not cropping at all.
    
    :param image: 
    :param crop_t: 
    :param crop_b: 
    :param crop_l: 
    :param crop_r: 
    :return: 
    """
    return 0, image.shape[0], 0, image.shape[1]


def skip_if_image_is_below_size(min_height=4, min_width=4):
    """
    Raises a Skip exception if the image size falls below the set image size.
    
    :param min_height: 
    :param min_width: 
    :return: 
    
    >>> skip_if_image_is_below_size(32, 32)(np.zeros((16,16)), Meta(0, 0))
    Traceback (most recent call last):
     ...
    mycelyso.pilyso.pipeline.executor.Skip: Meta(pos=0, t=<class 'mycelyso.pilyso.pipeline.executor.Collected'>)
    """
    def _inner(image, meta):
        if image.shape[0] < min_height or image.shape[1] < min_width:
            # noinspection PyCompatibility
            raise Skip(Meta(pos=meta.pos, t=Collected)) from None

        return image, meta

    return _inner


# noinspection PyUnusedLocal
def binarize(image, binary=None):
    """
    Binarizes the input image using the experimental thresholding technique.
    
    :param image: 
    :param binary: 
    :return: 
    """

    technique = getattr(binarization_module, ThresholdingTechnique.value)

    args, defaults = get_argnames_and_defaults(technique)

    for skip in 'kwargs', 'args':
        if skip in args:
            args.remove(skip)

    args = args[-len(defaults):]

    default_parameters = dict(zip(args, defaults))

    override_parameters = {}

    parameter_str = ThresholdingParameters.value
    if parameter_str:
        for pair in parameter_str.split(','):
            key, value = pair.split(':')

            if key not in default_parameters:
                raise RuntimeError("Unsupported parameter \"%s\" passed to thresholding function." % (key,))

            desired_type = type(default_parameters[key])

            override_parameters[key] = desired_type(value)

    return technique(image, **override_parameters)


# noinspection PyUnusedLocal
def skeletonize(binary, skeleton=None):
    """
    Skeletonizes the image using scikit-image's skeletonize function.
    
    :param binary: 
    :param skeleton: 
    :return:
    
    >>> skeletonize(np.array([[0, 0, 1, 1],
    ...                       [0, 0, 1, 1],
    ...                       [0, 0, 1, 1],
    ...                       [0, 0, 1, 1]]))
    array([[False, False, False, False],
           [False, False,  True, False],
           [False, False,  True, False],
           [False, False, False, False]], dtype=bool)
    """
    return sk_skeletonize(binary)


# noinspection PyUnusedLocal
def image_statistics(image, calibration, result=None):
    """
    Adds some numeric image parameters (i.e. size) to the results.
    
    :param image: 
    :param calibration: 
    :param result: 
    :return: 
    
    >>> sorted(image_statistics(np.array([[0, 0, 0],
    ...                                   [0, 0, 0],
    ...                                   [0, 0, 0]]), calibration=15.0).items())
    [('area', 2025.0), ('area_pixel', 9), ('input_height', 45.0), ('input_height_pixel', 3), \
('input_width', 45.0), ('input_width_pixel', 3)]
    """
    return {
        'input_width': image.shape[1] * calibration,
        'input_height': image.shape[0] * calibration,
        'input_width_pixel': image.shape[1],
        'input_height_pixel': image.shape[0],
        'area': image.shape[1] * calibration * image.shape[0] * calibration,
        'area_pixel': image.shape[1] * image.shape[0]
    }


# noinspection PyUnusedLocal
def quantify_binary(binary, calibration, result=None):
    """
    Adds some information about the binary image (i.e. covered ratio, area ...) to the results.
    
    :param binary: 
    :param calibration: 
    :param result: 
    :return: 
    
    >>> sorted(quantify_binary(np.array([[0, 0, 0],
    ...                                  [1, 1, 1],
    ...                                  [0, 0, 0]]), calibration=15.0).items())
    [('covered_area', 675.0), ('covered_area_pixel', 3), ('covered_ratio', 0.33333333333333331)]
    """
    ones = np.sum(binary)
    total = binary.shape[0] * binary.shape[1]
    return {
        'covered_ratio': ones / total,
        'covered_area': ones * calibration ** 2,
        'covered_area_pixel': ones
    }


# noinspection PyUnusedLocal
def graph_statistics(node_frame, result=None):
    """
    Adds some information about the graph to the results.
    
    :param node_frame: 
    :param result: 
    :return: 
    
    >>> pf = PixelFrame(np.array([[0, 0, 0],
    ...                           [1, 1, 1],
    ...                           [0, 0, 0]]), calibration=15.0)
    >>> sorted(graph_statistics(NodeFrame(pf)).items())
    [('graph_edge_count', 1.0), ('graph_edge_length', 30.0), ('graph_endpoint_count', 2), \
('graph_junction_count', 0), ('graph_node_count', 2)]
    """

    # everything / 2 because it's a digraph/graph structure
    graph_edge_length = node_frame.adjacency.sum() / 2
    graph_edge_count = node_frame.adjacency.nnz / 2

    graph_node_count = node_frame.adjacency.shape[0]
    graph_junction_count = len(node_frame.every_junction)
    graph_endpoint_count = len(node_frame.every_endpoint)

    return {
        'graph_edge_length': graph_edge_length * node_frame.calibration,
        'graph_edge_count': graph_edge_count,
        'graph_node_count': graph_node_count,
        'graph_junction_count': graph_junction_count,
        'graph_endpoint_count': graph_endpoint_count
    }


def clean_up(calibration, binary):
    """
    Cleans up the image by removing holes smaller than the configured size.
    
    :param calibration: 
    :param binary: 
    :return: 
    
    >>> clean_up(0.1, np.array([[ True,  True,  True],
    ...                         [ True, False,  True],
    ...                         [ True,  True,  True]]))
    array([[ True,  True,  True],
           [ True,  True,  True],
           [ True,  True,  True]], dtype=bool)
    """
    binary = (
        ndi.gaussian_filter(binary * 1.0, CleanUpGaussianSigma.value / calibration)
        > CleanUpGaussianThreshold.value
    )

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        binary = remove_small_holes(binary, min_size=int(CleanUpHoleFillSize.value / calibration ** 2), connectivity=2)

    return binary


def remove_small_structures(calibration, binary):
    """
    Cleans up the image by removing structures smaller than the configured size.
    
    :param calibration: 
    :param binary: 
    :return: 
    
    >>> remove_small_structures(0.1, np.array([[ False, False, False],
    ...                                        [ False, False,  True],
    ...                                        [ False, False,  True]]))
    array([[False, False, False],
           [False, False, False],
           [False, False, False]], dtype=bool)
    """
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        return remove_small_objects(binary,
                                    min_size=int(RemoveSmallStructuresSize.value / calibration ** 2),
                                    connectivity=2)  # TODO


def remove_border_artifacts(calibration, binary):
    """
    Removes structures, which are most likely artifacts because their centroid lies near the border.
    
    :param calibration: 
    :param binary: 
    :return: 
    
    >>> remove_border_artifacts(0.1, np.array([[ False, False, False],
    ...                                        [ False, False,  True],
    ...                                        [ False, False,  True]]))
    array([[False, False, False],
           [False, False, False],
           [False, False, False]], dtype=bool)
    """
    border = BorderArtifactRemovalBorderSize.value / calibration

    labeled = label(binary)
    corner_pixels = np.r_[
        labeled[0, :].ravel(),
        labeled[-1, :].ravel(),
        labeled[:, 0].ravel(),
        labeled[:, -1].ravel()
    ]

    corner_pixels = set(np.unique(corner_pixels)) - {0}

    for region in regionprops(labeled):
        if region.label in corner_pixels:
            if ((region.centroid[0] < border) or (region.centroid[0] > (binary.shape[0] - border)) or
                    (region.centroid[1] < border) or (region.centroid[1] > (binary.shape[1] - border))):
                binary[labeled == region.label] = False
    return binary


# noinspection PyUnusedLocal
def convert_to_nodes(skeleton, timepoint, calibration, pixel_frame=None, node_frame=None):
    """
    Passes the input skeleton into a PixelFrame and instantiates a NodeFrame based upon that.
    
    :param skeleton: 
    :param timepoint: 
    :param calibration: 
    :param pixel_frame: 
    :param node_frame: 
    :return: 
    
    """
    pixel_frame = PixelFrame(skeleton, timepoint, calibration=calibration)
    node_frame = NodeFrame(pixel_frame)
    return pixel_frame, node_frame


def track_multipoint(collected):
    """
    Initiates tracking between consecutive NodeFrames.
    
    :param collected: 
    :return: 
    """
    for result1, result2 in pairwise(collected.values()):
        result1.node_frame.track(result2.node_frame)
    return collected


# noinspection PyUnusedLocal
def individual_tracking(collected, tracked_fragments=None, tracked_fragments_fates=None):
    """
    After correspondence has been established by NodeFrame#track, reconstructs growing paths over time.
    
    :param collected: 
    :param tracked_fragments: 
    :param tracked_fragments_fates: 
    :return: 
    """
    def any_in(needles, haystack):
        for needle in needles:
            if needle in haystack:
                return True
        return False

    tracks = {}

    last_valid = []
    fates = {}

    next_track_id = 0

    for i, (current_result, next_result) in enumerate(pairwise(collected.values())):
        if current_result.covered_ratio > TrackingMaximumCoverage.value:
            break

        frame, next_frame = current_result.node_frame, next_result.node_frame

        calibration = frame.calibration

        timepoint, next_timepoint = frame.timepoint, next_frame.timepoint
        delta_t_in_hours = (next_timepoint - timepoint) / (60.0*60.0)

        # either: look for pairs of endpoints and junctions,
        # or endpoints and endpoints, if there are no junctions yet
        def valid_pairs():
            for _e in frame.every_endpoint:
                edges = frame.adjacency[_e, :]
                edge_indices = edges.nonzero()[1]

                for _other in edge_indices:
                    no_junctions = not any_in(frame.get_connected_nodes(_other), frame.every_junction)
                    if frame.is_junction(_other) or (no_junctions and frame.is_endpoint(_other)):
                        if _e != _other:
                            yield None, _e, _other

        valid = []
        endpoints_used = set()

        for track_id, e, other in chain(last_valid, valid_pairs()):
            e_on_next, other_on_next = frame.self_to_successor[e], frame.self_to_successor[other]

            if e_on_next == other_on_next:
                continue

            if e_on_next == -1 or other_on_next == -1:
                # tracking error
                fates[track_id] = "track aborted due to missing future node"
                continue

            distance = frame.shortest_paths[e, other]

            def distance_condition(current, new):
                if new == 0.0 or current == 0.0:
                    # print(new, current)
                    return False
                return (new > current) or (abs(1.0 - (current / new)) < TrackingMaximumRelativeShrinkage.value)

            if next_frame.connected_components[e_on_next] != next_frame.connected_components[other_on_next]:
                #  careful now: either the track broke, or we need to pick an alternative, fitting variant

                for _e_on_next, _other_on_next in product(frame.self_to_successor_alternatives[e],
                                                          frame.self_to_successor_alternatives[other]):
                    if (_e_on_next != _other_on_next and (
                        next_frame.connected_components[_e_on_next] == next_frame.connected_components[
                            _other_on_next]) and (
                        next_frame.shortest_paths[_e_on_next, _other_on_next] < float('inf')) and distance_condition(
                            distance, next_frame.shortest_paths[_e_on_next, _other_on_next])):
                        e_on_next, other_on_next = _e_on_next, _other_on_next
                        break

            next_distance = next_frame.shortest_paths[e_on_next, other_on_next]

            distance_num = frame.shortest_paths_num[e, other]

            # print(track_id, i, e, other, e_on_next, other_on_next, distance, distance_num)

            if e in endpoints_used or other in endpoints_used:
                fates[track_id] = "endpoints used otherwise"
                continue

            if distance == float('inf') or next_distance == float('inf'):
                # the first one should never happen,
                # the nodes are no longer connected on the next frame?
                fates[track_id] = "formerly connected components became unconnected? (dist/next dist %.4f %.4f)" % (
                    distance, next_distance
                )
                continue

            if ((next_distance < distance) and
                    abs(1.0 - (distance / next_distance)) > TrackingMaximumRelativeShrinkage.value):
                # a later distance was SHORTER than the current, that means tracking error or cycle in graph
                fates[track_id] = "track aborted due to shortcut (cycle or tracking error) [last %f > next %f]" % (
                    distance, next_distance
                )
                continue

            mu_per_h = ((next_distance - distance) * calibration) / delta_t_in_hours

            if not (TrackingMinimumTipElongationRate.value < mu_per_h < TrackingMaximumTipElongationRate.value):
                # a later distance changed too much, which might be a tracking error
                fates[track_id] = "track aborted due to too large change in length " \
                                  "[last %f, next %f, change %f mu per h]" % (
                    distance, next_distance, mu_per_h
                )
                continue

            # split up tracks which contain just endpoints ... otherwise the elongation rate is twice as high
            if frame.is_endpoint(e) and frame.is_endpoint(other):
                path = frame.get_path(e, other)
                if next_frame.is_endpoint(e_on_next) and next_frame.is_endpoint(other_on_next):
                    if len(path) > 2:
                        other = path[-2]
                        # noinspection PyUnusedLocal
                        distance = frame.shortest_paths[e, other]
                        # noinspection PyUnusedLocal
                        distance_num = frame.shortest_paths_num[e, other]
                        continue

            if track_id is None:
                track_id = next_track_id
                next_track_id += 1

            if track_id not in tracks:
                tracks[track_id] = {}

            tracks[track_id][i] = (e, other, e_on_next, other_on_next, distance, distance_num)

            valid.append((track_id, e_on_next, other_on_next))

            endpoints_used.add(e)

        last_valid = valid
    return tracks, fates


# noinspection PyProtectedMember,PyUnusedLocal
def prepare_tracked_fragments(collected, tracked_fragments, tracked_fragments_fates, track_table=None,
                              track_table_aux_tables=None):
    """
    Filters and converts tracked growing segments to result datasets.
    
    :param collected: 
    :param tracked_fragments: 
    :param tracked_fragments_fates: 
    :param track_table: 
    :param track_table_aux_tables: 
    :return: 
    """
    key_list = list(collected.keys())
    calibration = next(iter(collected.values()))['calibration']

    track_table = []

    track_table_aux_tables = []

    for track_id, track in tracked_fragments.items():
        if len(track) < TrackingMinimumTrackedPointCount.value:
            continue

        track_list = [[i, track[i]] for i in sorted(track.keys())]

        times_lengths = np.array(
            [[collected[key_list[i]]['node_frame'].timepoint, calibration * distance]
             for i, (e, other, e_on_next, other_on_next, distance, distance_num)
             in track_list])

        # minimum length of the maximally tracked segment
        if times_lengths[:, 1].max() < TrackingMinimalMaximumLength.value:
            continue

        # minimum length grown additionally
        if (times_lengths[:, 1].max() - times_lengths[:, 1].min()) < TrackingMinimalGrownLength.value:
            continue

        # whatever the settings are, there must be growth
        if (times_lengths[:, 1].max() - times_lengths[:, 1].min()) == 0.0:
            continue

        aux_table_num = len(track_table_aux_tables)

        track_table_aux_tables.append(
            [{
                 'track_table_number': aux_table_num,
                 'timepoint': collected[key_list[i]]['node_frame'].timepoint,
                 'node_id_a': e,
                 'node_id_b': other,
                 'node_next_id_a': e_on_next,
                 'node_next_id_b': other_on_next,
                 'distance': calibration * distance,
                 'distance_num': distance_num
             } for i, (e, other, e_on_next, other_on_next, distance, distance_num) in track_list]
        )

        distance_num_helper = np.array(
            [distance_num for i, (e, other, e_on_next, other_on_next, distance, distance_num) in track_list])

        times = times_lengths[:, 0]
        lengths = times_lengths[:, 1]
        relative_lengths = times_lengths[:, 1] / times_lengths[:, 1].min()

        t_min, t_max = times.min(), times.max()

        row = {
            'timepoint_begin': t_min,
            'timepoint_end': t_max,
            'timepoint_center': t_min + (t_max - t_min) / 2,
            'minimum_distance': lengths.min(),
            'maximum_distance': lengths.max(),
            'minimum_distance_num': distance_num_helper.min(),
            'maximum_distance_num': distance_num_helper.max(),
            'duration': t_max - t_min,
            'count': len(times),
            'aux_table': aux_table_num
        }

        try:

            regression = linregress(times, lengths)._asdict()
            row.update({'plain_regression_' + k: v for k, v in regression.items()})
            regression = linregress(times, relative_lengths)._asdict()
            row.update({'normalized_regression_' + k: v for k, v in regression.items()})

            regression = linregress(times, np.log(lengths))._asdict()
            row.update({'logarithmic_plain_regression_' + k: v for k, v in regression.items()})
            regression = linregress(times, np.log(relative_lengths))._asdict()
            row.update({'logarithmic_normalized_regression_' + k: v for k, v in regression.items()})

            regression = prepare_optimized_regression(times, lengths)
            row.update({'optimized_regression_' + k: v for k, v in regression.items()})

            regression = prepare_optimized_regression(times, relative_lengths)
            row.update({'optimized_normalized_regression_' + k: v for k, v in regression.items()})

            regression = prepare_optimized_regression(times, np.log(lengths))
            row.update({'optimized_logarithmic_regression_' + k: v for k, v in regression.items()})

            regression = prepare_optimized_regression(times, np.log(relative_lengths))
            row.update({'optimized_logarithmic_normalized_regression_' + k: v for k, v in regression.items()})
        except IndexError:
            pass

        track_table.append(row)

    return track_table, track_table_aux_tables


def prepare_position_regressions(collected, result):
    """
    Prepares some regressions over parameters collected per position over time.
    
    :param collected: 
    :param result: 
    :return: 
    """
    fields = ['covered_ratio', 'covered_area', 'graph_edge_length', 'graph_edge_count', 'graph_node_count',
              'graph_junction_count', 'graph_endpoint_count']

    row = {}

    # noinspection PyProtectedMember
    def prepare_for_field(field_name):
        data = np.array([[f.timepoint, f[field_name]] for f in collected.values()])

        regression = linregress(data[:, 0], data[:, 1])._asdict()
        row.update({field_name + '_linear_regression_' + k: v for k, v in regression.items()})

        regression = linregress(data[:, 0], np.log(data[:, 1]))._asdict()
        row.update({field_name + '_logarithmic_regression_' + k: v for k, v in regression.items()})

        regression = prepare_optimized_regression(data[:, 0], data[:, 1])
        row.update({field_name + '_optimized_linear_regression_' + k: v for k, v in regression.items()})

        regression = prepare_optimized_regression(data[:, 0], np.log(data[:, 1]))
        row.update({field_name + '_optimized_logarithmic_regression_' + k: v for k, v in regression.items()})

    for field in fields:
        try:
            with np.errstate(divide='ignore', invalid='ignore'):
                prepare_for_field(field)
        except IndexError:
            pass

    result.update(row)

    return result


# noinspection PyUnusedLocal
def generate_graphml(node_frame, result):
    """
    Generates a GraphML representation of a particular frame.
    
    :param node_frame: 
    :param result: 
    :return: 
    """
    return {
        'graphml': to_graphml_string(node_frame.get_networkx_graph())
    }


# noinspection PyTypeChecker,PyUnusedLocal
def generate_overall_graphml(collected, result):
    """
    Generates a GraphML representation of the whole graph of one image stack.
    
    :param collected: 
    :param result: 
    :return: 
    """
    time_to_z_scale = 1.0

    graphs_list = {}
    successors = {}
    node_counts = {}
    node_count_accumulator = 0

    graph = nx.Graph()

    for i, (meta, frame) in enumerate(list(collected.items())[:-1]):
        node_frame = frame.node_frame
        successors[i] = node_frame.self_to_successor
        g = node_frame.get_networkx_graph(with_z=time_to_z_scale)
        g = nx.relabel_nodes(g, lambda node_id: node_id + node_count_accumulator, copy=True)
        graphs_list[i] = g
        node_counts[i] = node_count_accumulator
        node_count_accumulator += len(g.nodes())
        graph.add_nodes_from(g.nodes(data=True))
        graph.add_edges_from(g.edges(data=True))

    for i, mapping in successors.items():
        if i + 1 == len(successors):
            break
        for relative_from_index, relative_to_index in enumerate(mapping):
            if relative_to_index == -1:
                continue
            from_index = node_counts[i] + relative_from_index
            to_index = node_counts[i + 1] + relative_to_index
            graph.add_edge(from_index, to_index)

    return {
        'overall_graphml': to_graphml_string(graph)
    }



# -*- coding: utf-8 -*-
"""
The ND2 reader module implements an ImageStack able to open Nikon ND2 images, using the nd2file module.
"""

from ..imagestack import ImageStack, Dimensions

from nd2file import ND2MultiDim


class ND2ImageStack(ImageStack):
    extensions = ('.nd2',)

    priority = 500

    def open(self, location, **kwargs):
        self.nd2 = ND2MultiDim(location.path)

        self.set_dimensions_and_sizes(
            [Dimensions.Time, Dimensions.PositionXY, Dimensions.PositionZ, Dimensions.Channel],
            [int(self.nd2.timepointcount), int(self.nd2.multipointcount), 1, int(self.nd2.channels)]
        )

    def get_data(self, what):

        image = self.nd2.image(multipoint=what[Dimensions.PositionXY], timepoint=what[Dimensions.Time])

        if len(image.shape) == 3:
            try:
                channel = what[Dimensions.Channel]
            except KeyError:
                channel = 0

            image = image[:, :, channel]

        return image

    def get_meta(self, what):

        nan = float('nan')

        pos_meta = {'x': nan, 'y': nan, 'z': nan}

        try:
            pos_meta.update(self.nd2.multipoints[what[Dimensions.PositionXY]])
        except KeyError:
            pass

        position = self.__class__.Position(x=float(pos_meta['x']), y=float(pos_meta['y']), z=float(pos_meta['z']))

        meta = self.__class__.Metadata(
            time=float(
                self.nd2.get_time(
                    self.nd2.calc_num(multipoint=what[Dimensions.PositionXY], timepoint=what[Dimensions.Time])
                )
            ),
            position=position,
            calibration=self.nd2.calibration)

        return meta




# -*- coding: utf-8 -*-
"""
The test module's __main__ contains the main() function to run the doctests.
"""

import sys
import doctest


# noinspection PyUnresolvedReferences
def main():
    """
    Runs all the doctests.

    """

    import mycelyso.highlevel.nodeframe
    import mycelyso.highlevel.pipeline
    import mycelyso.highlevel.pixelframe
    import mycelyso.highlevel.steps

    import mycelyso.misc.graphml
    import mycelyso.misc.regression
    import mycelyso.misc.util

    import mycelyso.pilyso.application.application
    import mycelyso.pilyso.application.helper

    import mycelyso.pilyso.imagestack.imagestack

    import mycelyso.pilyso.misc.h5writer
    import mycelyso.pilyso.misc.processpool

    import mycelyso.pilyso.pipeline.executor
    import mycelyso.pilyso.pipeline.pipeline

    import mycelyso.pilyso.steps.steps

    import mycelyso.processing.binarization
    import mycelyso.processing.pixelgraphs

    modules_to_test = [
        mycelyso.highlevel.nodeframe,
        mycelyso.highlevel.pipeline,
        mycelyso.highlevel.pixelframe,
        mycelyso.highlevel.steps,
        #
        mycelyso.misc.graphml,
        mycelyso.misc.regression,
        mycelyso.misc.util,
        #
        mycelyso.pilyso.application.application,
        mycelyso.pilyso.application.helper,
        #
        mycelyso.pilyso.imagestack.imagestack,
        #
        mycelyso.pilyso.misc.h5writer,
        mycelyso.pilyso.misc.processpool,
        #
        mycelyso.pilyso.pipeline.executor,
        mycelyso.pilyso.pipeline.pipeline,
        #
        mycelyso.pilyso.steps.steps,
        #
        mycelyso.processing.binarization,
        mycelyso.processing.pixelgraphs,
    ]

    total_failures, total_tests = 0, 0

    for a_module in modules_to_test:
        failures, tests = doctest.testmod(a_module)
        total_failures += failures
        total_tests += tests

    print("Run %d tests in total." % (total_tests,))

    if total_failures > 0:
        print("Test failures occurred, exiting with non-zero status.")
        sys.exit(1)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
The pipeline module contains the mycelyso-Pipeline, assembled from various functions.
"""

from os.path import basename, abspath
# noinspection PyUnresolvedReferences
import networkx as nx

from tunable import TunableManager

from ..tunables import CropWidth, CropHeight, BoxDetection, StoreImage, SkipBinarization
from .steps import *
from .. import __banner__
from .. import __version__
# noinspection PyUnresolvedReferences
from ..pilyso.application import App, PipelineExecutionContext, PipelineEnvironment, Every, Collected, Meta, Skip
from ..pilyso.imagestack import ImageStack
from ..pilyso.misc.h5writer import hdf5_output, hdf5_node_name, return_or_uncompress
from ..pilyso.pipeline.pipeline import NeatDict
from ..pilyso.steps import \
    image_source, pull_metadata_from_image, substract_start_frame, rescale_image_to_uint8, set_result, Delete, \
    box_detection, create_boxcrop_from_subtracted_image, calculate_image_sha256_hash, Compress


class Mycelyso(App):
    """
    The Mycelyso App, implementing a pilyso App.
    """
    def options(self):
        return {
            'name': "mycelyso",
            'description': "",
            'banner': __banner__,
            'pipeline': MycelysoPipeline
        }

    def arguments(self, argparser):
        argparser.add_argument('--meta', '--meta', dest='meta', default='')
        argparser.add_argument('--interactive', '--interactive', dest='interactive',
                               default=False, action='store_true')
        argparser.add_argument('--output', '--output', dest='output', default='output.h5')

    def handle_args(self):
        self.args.tunables = TunableManager.get_representation()

        if self.args.interactive:
            # if interactive, don't spawn workers
            self.args.processes = 0
            self.run = self.interactive_run

    def interactive_run(self):
        pipeline, fun, args, kwargs = self.pe.complete_args
        assert fun == '__init__'
        pipeline = pipeline(*args, **kwargs)

        import matplotlib.pyplot as plt
        from matplotlib.widgets import Slider

        fig, ax = plt.subplots()

        plt.subplots_adjust(left=0.25, bottom=0.25)

        fig.canvas.set_window_title("Image Viewer")

        slider_background = '#e7af12'
        slider_foreground = '#005b82'

        ax_mp = plt.axes([0.25, 0.1, 0.65, 0.03], facecolor=slider_background)
        ax_tp = plt.axes([0.25, 0.15, 0.65, 0.03], facecolor=slider_background)

        mp_max = max(self.positions)
        tp_max = max(self.timepoints)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            multipoint = Slider(ax_mp, 'Multipoint', 0, mp_max, valinit=0, valfmt="%d", color=slider_foreground)
            timepoint = Slider(ax_tp, 'Timepoint', 0, tp_max, valinit=0, valfmt="%d", color=slider_foreground)

        env = {'show': True}

        def update(_):
            t = int(timepoint.val)
            position = int(multipoint.val)

            fig.canvas.set_window_title("Image Viewer - [BUSY]")

            plt.rcParams['image.cmap'] = 'gray'

            plt.sca(ax)
            plt.cla()

            plt.suptitle('[left/right] timepoint [up/down] multipoint [h] hide analysis')

            result = pipeline.dispatch(Meta(pos=Every, t=Every), meta=Meta(pos=position, t=t))

            result = NeatDict(result)

            binary = return_or_uncompress(result.binary)
            skeleton = return_or_uncompress(result.skeleton)

            node_draw_parameters = dict(node_size=25, node_color='darkgray', linewidths=0, edge_color='darkgray')

            graph, pos = result.node_frame.get_networkx_graph(return_positions=True)

            nx.draw_networkx(
                graph,
                with_labels=False,
                pos=pos,
                **node_draw_parameters
            )

            if env['show']:
                plt.imshow(binary, cmap='gray_r')
            else:
                plt.imshow(skeleton, cmap='gray_r')

            fig.canvas.set_window_title("Image Viewer - %s timepoint #%d %d/%d multipoint #%d %d/%d" %
                                        (self.args.input, t, 1 + t, 1 + tp_max, position, 1 + position, 1 + mp_max))

            plt.draw()

        update(None)

        multipoint.on_changed(update)
        timepoint.on_changed(update)

        def key_press(event):
            if event.key == 'left':
                timepoint.set_val(max(1, int(timepoint.val) - 1))
            elif event.key == 'right':
                timepoint.set_val(min(tp_max, int(timepoint.val) + 1))
            elif event.key == 'ctrl+left':
                timepoint.set_val(max(1, int(timepoint.val) - 10))
            elif event.key == 'ctrl+right':
                timepoint.set_val(min(tp_max, int(timepoint.val) + 10))
            elif event.key == 'down':
                multipoint.set_val(max(1, int(multipoint.val) - 1))
            elif event.key == 'up':
                multipoint.set_val(min(mp_max, int(multipoint.val) + 1))
            elif event.key == 'h':
                env['show'] = not env['show']
                update(None)
            elif event.key == 'q':
                raise SystemExit

        fig.canvas.mpl_connect('key_press_event', key_press)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            fig.tight_layout()

        plt.show()


class MycelysoPipeline(PipelineExecutionContext):
    """
    The MycelysoPipeline, defining the pipeline (with slight alterations based upon arguments passed via command line).
    """
    def __init__(self, args):
        TunableManager.load(args.tunables)

        absolute_input = abspath(args.input)
        h5nodename = hdf5_node_name(absolute_input)

        self.pipeline_environment = PipelineEnvironment(ims=ImageStack(args.input))

        per_image = self.add_stage(Meta(pos=Every, t=Every))

        per_image |= set_result(tunables_hash=TunableManager.get_hash())

        # read the image
        per_image |= image_source
        per_image |= calculate_image_sha256_hash
        per_image |= pull_metadata_from_image

        per_image |= lambda image, raw_image=None: image
        per_image |= lambda image, raw_unrotated_image=None: image

        per_image |= set_empty_crops

        # define what we want (per image) as results

        result_table = {
            '_plain': [
                'calibration', 'timepoint', 'input_height',
                'input_width', 'area', 'covered_ratio', 'covered_area',
                'graph_edge_length', 'graph_edge_count', 'graph_node_count',
                'graph_junction_count', 'graph_endpoint_count',
                'filename', 'metadata', 'shift_x', 'shift_y',
                'crop_t', 'crop_b', 'crop_l', 'crop_r',
                'image_sha256_hash', 'tunables_hash'
            ],
            'graphml': 'data',
            # 'image': 'image',
            # 'raw_unrotated_image': 'image',
            # 'raw_image': 'image',
            'skeleton': 'image',
            'binary': 'image'
        }

        if StoreImage.value:
            result_table['image'] = 'image'

        per_image |= set_result(
            reference_timepoint=1,
            filename_complete=absolute_input,
            filename=basename(absolute_input),
            metadata=args.meta,
            result_table=result_table
        )

        per_image |= substract_start_frame

        if BoxDetection.value:
            per_image |= box_detection
            per_image |= create_boxcrop_from_subtracted_image

        per_image |= rescale_image_to_uint8

        per_image |= set_result(raw_unrotated_image=Delete, raw_image=Delete, subtracted_image=Delete)

        per_image |= lambda image: image[
                                   CropHeight.value:-(CropHeight.value if CropHeight.value > 0 else 1),
                                   CropWidth.value:-(CropWidth.value if CropWidth.value > 0 else 1)
                                   ]

        per_image |= lambda crop_t, crop_b, crop_l, crop_r: (
            crop_t + CropHeight.value,
            crop_b - CropHeight.value,
            crop_l + CropWidth.value,
            crop_r - CropWidth.value
        )

        per_image |= skip_if_image_is_below_size(4, 4)

        # generate statistics of the image
        per_image |= image_statistics

        if not SkipBinarization.value:
            # binarize
            per_image |= binarize
        else:
            # noinspection PyUnusedLocal
            def _image_to_binary(image, binary=None):
                return image.astype(bool)

            per_image |= _image_to_binary

        # ... and cleanup
        per_image |= clean_up

        per_image |= remove_small_structures

        per_image |= remove_border_artifacts

        # generate statistics of the binarized image
        per_image |= quantify_binary

        per_image |= skeletonize

        # 'binary', 'skeleton' are kept!
        per_image |= convert_to_nodes

        if not StoreImage.value:
            per_image |= set_result(image=Delete)

        per_image |= set_result(pixel_frame=Delete)

        per_image |= graph_statistics
        per_image |= generate_graphml

        per_image |= set_result(binary=Compress, skeleton=Compress, graphml=Compress)

        # per position

        per_position = self.add_stage(Meta(pos=Every, t=Collected))

        per_position |= track_multipoint

        per_position |= generate_overall_graphml

        per_position |= individual_tracking

        per_position |= prepare_tracked_fragments

        per_position |= prepare_position_regressions

        per_position |= lambda meta, meta_pos=None: meta.pos

        per_position |= set_result(
            filename_complete=absolute_input,
            filename=basename(absolute_input),
            metadata=args.meta,
            tunables=TunableManager.get_serialization(),
            version=__version__,
            banner=__banner__,
            result_table={
                '_plain': [
                    'metadata',
                    'filename_complete',
                    'filename',
                    'meta_pos',
                    '*_regression_*'
                ],
                'tunables': 'data',
                'version': 'data',
                'banner': 'data',
                'overall_graphml': 'data',
                'track_table': 'table',
                'track_table_aux_tables': 'table'
            }
        )

        per_position |= hdf5_output(args.output, h5nodename)

        def black_hole(result):
            for k in list(result.keys()):
                del result[k]
            del result
            return {}

        per_position |= black_hole



# -*- coding: utf-8 -*-
"""
The binarization module contains the binarization routine used to segment phase contrast images of 
mycelium networks into foreground and background.
"""

import numpy as np

try:
    import numexpr
except ImportError:
    numexpr = None

from skimage.transform import integral_image
from skimage.feature import shape_index
from skimage.filters import rank


def mean_and_std(image, window_size=15):
    """
    Helper function returning mean and average images sped up using integral images / summed area tables.

    :param image: Input image
    :param window_size: Window size
    :return: tuple (mean, std)
    """

    enlarged = np.pad(image.astype(np.float64), window_size, 'reflect')

    def integral_image_and_squared(image):
        cv2 = None  # could be sped up using OpenCV
        if cv2:
            sums_, sqsums_ = cv2.integral2(image)
            return sums_[1:, 1:], sqsums_[1:, 1:]
        else:
            return integral_image(image), integral_image(image ** 2)

    ints, ints_sq = integral_image_and_squared(enlarged)

    def calculate_sums(mat):
        a = mat[:-2 * window_size, :-2 * window_size]
        b = mat[2 * window_size:, 2 * window_size:]
        c = mat[:-2 * window_size, 2 * window_size:]
        d = mat[2 * window_size:, :-2 * window_size]
        if numexpr:
            return numexpr.evaluate("(a + b) - (c + d)").astype(np.float32)
        else:
            return (a + b) - (c + d)

    sums = calculate_sums(ints)
    sums_squared = calculate_sums(ints_sq)

    area = (2.0 * window_size + 1) ** 2

    mean = sums / area

    if numexpr:
        std = numexpr.evaluate("sqrt(sums_squared / area - mean ** 2)")
    else:
        std = np.sqrt(sums_squared / area - mean ** 2)

    return mean, std


def normalize(image):
    """
    Normalizes an image to the range 0-1

    :param image:
    :return:
    """

    image = image.astype(np.float32)
    image -= image.min()
    image /= image.max()

    return image


# noinspection PyUnusedLocal
def sauvola(image, window_size=15, k=0.5, r=128, return_threshold=False, **kwargs):
    """
    Thresholding method as developed by [Sauvola1997]_.

    .. [Sauvola1997] Sauvola et al. (1997) Proc. Doc. Anal. Recog.
       DOI: `10.1109/ICDAR.1997.619831 <https://dx.doi.org/10.1109/ICDAR.1997.619831>`_

    :param image: Input image
    :param window_size: Window size
    :param k: k value
    :param r: r value
    :param return_threshold: Whether to return a binarization, or the actual threshold values
    :param kwargs: For compatibility
    :return:
    """

    mean, std = mean_and_std(image, window_size)

    if numexpr:
        threshold = numexpr.evaluate("mean * (1.0 + k * ((std / r) - 1.0))")
    else:
        threshold = mean * (1.0 + k * ((std / r) - 1.0))

    if return_threshold:
        return threshold
    else:
        return image < threshold


# noinspection PyUnusedLocal
def phansalkar(image, window_size=15, k=0.25, r=0.5, p=2.0, q=10.0, return_threshold=False, **kwargs):
    """
    Thresholding method as developed by [Phansalkar2011]_.

    .. [Phansalkar2011] Phansalkar et al. (2011) Proc. ICCSP
       DOI: `10.1109/ICCSP.2011.5739305 <https://dx.doi.org/10.1109/ICCSP.2011.5739305>`_

    :param image: Input image
    :param window_size: Window size
    :param k: k value
    :param r: r value
    :param p: p value
    :param q: q value
    :param return_threshold: Whether to return a binarization, or the actual threshold values
    :param kwargs: For compatibility
    :return:
    """

    original_image = image.copy()
    image = normalize(image)
    mean, std = mean_and_std(image, window_size)

    if numexpr:
        threshold = numexpr.evaluate("mean * (1.0 + p * exp(-q * mean) + k * ((std / r) - 1.0))")
    else:
        threshold = mean * (1.0 + p * np.exp(-q * mean) + k * ((std / r) - 1.0))

    if return_threshold:
        image_range = original_image.max()-original_image.min()
        return threshold * image_range + original_image.min()
    else:
        return image < threshold


# noinspection PyUnusedLocal,PyPep8Naming
def wolf(image, mask=None, window_size=15, a=0.5, return_threshold=False, **kwargs):
    """
    Thresholding method as developed by [Wolf2004]_.

    .. [Wolf2004] Wolf & Jolion (2004) Form. Pattern Anal. & App.
       DOI: `10.1007/s10044-003-0197-7 <https://dx.doi.org/10.1007/s10044-003-0197-7>`_

    :param image: Input image
    :param mask: Possible mask denoting a ROI
    :param window_size: Window size
    :param a: a value
    :param return_threshold: Whether to return a binarization, or the actual threshold values
    :param kwargs: For compatibility
    :return:
    """

    mean, std = mean_and_std(image, window_size)

    if mask is not None:
        M = image[mask].min()
        R = std[mask].max()
    else:
        M = image.min()
        R = std.max()

    if numexpr:
        threshold = numexpr.evaluate("(1-a) * mean + a*M + a*(std/R)*(mean - M)")
    else:
        threshold = (1-a) * mean + a*M + a*(std/R)*(mean - M)

    if return_threshold:
        return threshold
    else:
        return image < threshold


# noinspection PyUnusedLocal,PyUnusedLocal,PyPep8Naming
def feng(image, mask=None,
         window_size=15, window_size2=30,
         a1=0.12, gamma=2, k1=0.25, k2=0.04,
         return_threshold=False, **kwargs):
    """
    Thresholding method as developed by [Feng2004]_.

    .. [Feng2004] Fend & Tan (2004) IEICE Electronics Express
       DOI: `10.1587/elex.1.501 <https://dx.doi.org/10.1587/elex.1.501>`_

    :param image: Input image
    :param mask: Possible mask denoting a ROI
    :param window_size: Window size
    :param window_size2: Second window size
    :param a1: a1 value
    :param gamma: gamma value
    :param k1: k1 value
    :param k2: k2 value
    :param return_threshold: Whether to return a binarization, or the actual threshold values
    :param kwargs: For compatibility
    :return:
    """

    mean, std = mean_and_std(image, window_size)

    mean2, std2 = mean_and_std(image, window_size2)

    M = rank.minimum(image, np.ones((window_size, window_size)))
    # what exactly do they mean? maximum of window?
    Rs = rank.maximum(std2.astype(np.uint8), np.ones((window_size2, window_size2)))

    if numexpr:
        threshold = numexpr.evaluate("""(
        (1 - a1) * mean + (k1 * (std / Rs) ** gamma) * (std/Rs) * (mean - M) + (k2 * (std / Rs) ** gamma) * M
        )""")
    else:
        a2 = k1 * (std / Rs) ** gamma
        a3 = k2 * (std / Rs) ** gamma
        threshold = (1 - a1) * mean + a2 * (std/Rs) * (mean - M) + a3 * M

    if return_threshold:
        return threshold
    else:
        return image < threshold


# noinspection PyUnusedLocal
def nick(image, window_size=15, k=-0.1, return_threshold=False, **kwargs):
    """
    Thresholding method as developed by [Khurshid2009]_.

    .. [Khurshid2009] Khurshid et al. (2009) Proc. SPIE
       DOI: `10.1117/12.805827 <https://dx.doi.org/10.1117/12.805827>`_

    :param image: Input image
    :param window_size: Window size
    :param k: k value
    :param return_threshold: Whether to return a binarization, or the actual threshold values
    :param kwargs: For compatibility
    :return:
    """
    mean, std = mean_and_std(image, window_size)

    if numexpr:
        image_size = image.size
        image_sq_sum = (image**2).sum()
        threshold = numexpr.evaluate("mean + k * sqrt((image_sq_sum - mean**2)/image_size)")
    else:
        threshold = mean + k * np.sqrt(((image**2).sum() - mean**2)/image.size)

    if return_threshold:
        return threshold
    else:
        return image < threshold


# noinspection PyUnusedLocal
def bataineh(image, mask=None, window_size=15, return_threshold=False, **kwargs):
    """
    Thresholding method as developed by [Bataineh2011a]_.

    .. [Bataineh2011a] Bataineh et al. (2011) Pattern Recognit. Lett.
       DOI: `10.1016/j.patrec.2011.08.001 <https://dx.doi.org/10.1016/j.patrec.2011.08.001>`_

    :param image: Input image
    :param mask: Possible mask denoting a ROI
    :param window_size: Window size
    :param return_threshold: Whether to return a binarization, or the actual threshold values
    :param kwargs: For compatibility
    :return:
    """

    mean, std = mean_and_std(image, window_size)
    if mask is not None:
        adaptive_stddev = (std - std[mask].min()) / (std[mask].max() - std[mask].min())
        image_mean = image[mask].mean()
    else:
        adaptive_stddev = (std - std.min()) / (std.max() - std.min())
        image_mean = image.mean()

    if numexpr:
        threshold = numexpr.evaluate("mean - ((mean**2.0 - std) / ((image_mean + std) * (std + adaptive_stddev)))")
    else:
        threshold = mean - ((mean**2.0 - std) / ((image_mean + std) * (std + adaptive_stddev)))

    if return_threshold:
        return threshold
    else:
        return image < threshold


# I had highly optimized versions of sub functions of these, mainly
# using cv2, however to keep installation easy, I've replaced them by slower scikit-image etc. functions

# noinspection PyUnusedLocal
def experimental_thresholding(image, mask=None, window_size=15,
                              gaussian_sigma=3.0, shift=0.2, target=-0.5, quotient=1.2,
                              return_threshold=False, **kwargs):
    """
    A novel thresholding method basing upon the shape index as defined by [Koenderink1992]_, and [Bataineh2011]_
    automatic adaptive thresholding. The method is due to be explained in detail in the future.
    
    .. [Koenderink1992] Koenderink and van Doorn (1992) Image Vision Comput.
       DOI: `10.1016/0262-8856(92)90076-F <https://dx.doi.org/10.1016/0262-8856(92)90076-F>`_
    .. [Bataineh2011] Bataineh et al. (2011) Pattern Recognit. Lett.
       DOI: `10.1016/j.patrec.2011.08.001 <https://dx.doi.org/10.1016/j.patrec.2011.08.001>`_
      
    
    :param image: Input image
    :param mask: Possible mask denoting a ROI
    :param window_size: Window size
    :param gaussian_sigma: Sigma of the Gaussian used for smoothing
    :param shift: Shift parameter
    :param target: Target shape index parameter
    :param quotient: Quotient parameter
    :param return_threshold: Whether to return a binarization, or the actual threshold values
    :param kwargs: For compatibility
    :return: 
    """
    # novel method based upon shape index and Bataineh thresholding

    means, stddev = mean_and_std(image, window_size)

    with np.errstate(invalid='ignore'):
        sim = shape_index(image, gaussian_sigma)

    if mask is not None:
        adaptive_stddev = (stddev - stddev[mask].min()) / (stddev[mask].max() - stddev[mask].min())
        image_mean = image[mask].mean()
    else:
        adaptive_stddev = (stddev - stddev.min()) / (stddev.max() - stddev.min())
        image_mean = image.mean()

    if numexpr:
        threshold = numexpr.evaluate(
            "(exp((-(sim - target)**2)/quotient) + shift)*"
            "means*"
            "((image_mean + stddev) * (stddev + adaptive_stddev))/(means**2 - stddev)"
        )
    else:
        threshold = (
            (np.exp((-(sim - target)**2)/quotient) + shift) *
            means *
            ((image_mean + stddev) * (stddev + adaptive_stddev))/(means**2 - stddev)
        )

    if return_threshold:
        return threshold
    else:
        return image < threshold



# -*- coding: utf-8 -*-
"""
pilyso ... python image analysis software

This library has devised to serve as base for multiple image analysis tools.
So far it is not completely ready as a stand-alone library,
and as such just bundled with mycelyso as-of now.
"""



# -*- coding: utf-8 -*-
"""
The recursionlimit_raise submodule raises the recursion limit of the Python interpreter upon import.
(Currently to 10**9)
"""

import sys
# graphs, recursion, let's raise the bar a bit ;)
sys.setrecursionlimit(int(1E9))




# encoding: utf-8

from collections import OrderedDict
from inspect import isclass
import inspect

if getattr(inspect, 'signature'):
    def get_argnames_and_defaults(call):
        sig = inspect.signature(call)
        args = [para for para in sig.parameters]
        # noinspection PyProtectedMember
        defaults = [para.default for para in sig.parameters.values() if para.default is not inspect._empty]
        return args, defaults

else:
    def get_argnames_and_defaults(call):
        # noinspection PyDeprecation
        argspec = inspect.getargspec(call)

        args = list(argspec.args)
        defaults = argspec.defaults if list(argspec.defaults) else []

        return args, defaults


class NeatDict(dict):
    def __getattr__(self, item):
        return self.get(item)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        del self[item]


class Pipeline(object):

    def __init__(self):
        self.steps = []

    def __lt__(self, other):
        return self.__le__(other)

    def __le__(self, other):
        return self.call(other)

    def __or__(self, other):
        self.add(other)
        return self

    def __ior__(self, other):
        self.add(other)
        return self

    def add(self, what):
        self.steps.append(what)

    def call(self, parameter):
        return PipelineInvocation(self, parameter)


class PipelineInvocation(object):
    def __init__(self, pipeline, parameter):
        self.pipeline = pipeline
        self.parameter = parameter


class PipelineEnvironment(object):
    debug = False

    KEY_COLLECTED = 'collected'
    KEY_RESULT = 'result'

    def __init__(self, **kwargs):
        self.external_di = {'pipeline_environment': self}
        self.external_di.update(kwargs)

    @staticmethod
    def new_pipeline():
        return Pipeline()

    def run(self, pipeline_invocation, **kwargs):

        if type(pipeline_invocation) == Pipeline:
            pipeline_invocation = pipeline_invocation < kwargs

        result = pipeline_invocation.parameter
        for raw_step in pipeline_invocation.pipeline.steps:
            step = self.wrap(raw_step)

            result = step(result)
        return result

    def debug_message(self, *args, **kwargs):
        if self.debug:
            print(args, kwargs)
        else:
            pass

    def wrap_result(self, result):

        result = NeatDict(result)

        result[self.KEY_RESULT] = True

        if self.KEY_COLLECTED in result:
            wrapped = OrderedDict()
            for k, v in result[self.KEY_COLLECTED].items():
                if v is None:
                    # wrapped[k] = None
                    pass
                else:
                    wrapped[k] = NeatDict(v)
            result[self.KEY_COLLECTED] = wrapped

        return result

    def unwrap_result(self, result):
        if self.KEY_COLLECTED in result and (result[self.KEY_COLLECTED] is not None):
            unwrapped = OrderedDict()
            for k, v in result[self.KEY_COLLECTED].items():
                unwrapped[k] = dict(v)
            result[self.KEY_COLLECTED] = unwrapped

        result = dict(result)

        if self.KEY_RESULT in result:
            del result[self.KEY_RESULT]

        return result

    def prepare_call(self, call, di=None, result=None):
        if di is None:
            di = {}

        if result is None:
            result = {}

        args, defaults = get_argnames_and_defaults(call)

        if args[0] == 'self':
            args = args[1:]

        non_default_parameters = len(args) - len(defaults)

        def _wrapped():
            from_di = set()
            parameters = []

            for n, arg in enumerate(args):
                if arg == self.KEY_RESULT:
                    parameters.append(result)
                elif arg in result:
                    parameters.append(result[arg])
                else:
                    if n >= non_default_parameters:
                        parameters.append(defaults[n - non_default_parameters])
                    else:
                        if arg in di:
                            from_di.add(arg)
                            parameters.append(di[arg])
                        else:
                            # problem: pipeline step asks for a parameter we do not have
                            raise ValueError('[At %s]: Argument %r not in %r' % (repr(call), arg, result,))

            _call_return = call(*parameters)
            return args, _call_return

        return _wrapped

    def reassemble_result(self, result, args, _call_return):
        if type(_call_return) == dict and self.KEY_RESULT in _call_return:
            # if we get a dict back, we merge it with the ongoing result object
            result.update(_call_return)
        elif type(_call_return) == NeatDict:
            # if we get a neatdict back, we assume its the proper result object
            # and the pipeline step knew what it did ...
            # we continue with it as-is
            result = _call_return
        else:
            if type(_call_return) != tuple:
                _call_return = (_call_return,)

            for n, item in enumerate(reversed(_call_return)):
                k = args[-(n+1)]
                if k == self.KEY_RESULT:
                    if type(item) == dict or type(item) == NeatDict:
                        result.update(item)
                else:
                    # do nothing if the parameter came from di
                    if k not in self.external_di:
                        result[k] = item
        return result

    def wrap(self, what):
        name = ("class_" if isclass(what) else "function_") +\
            ('__LAMBDA__' if getattr(what, '__name__', False) == '<lambda>' else getattr(what, '__name__', repr(what)))

        if isclass(what):
            # we create an instance without calling the constructor, so we can setup the environment first
            instance = what.__new__(what)
            # now we call the constructor, and it has everything neatly set up already!

            init_call = self.prepare_call(instance.__init__, di=self.external_di)
            init_call()

            call = instance
        else:
            call = what

        def _wrapped(result):
            self.debug_message("Entering %s", name)
            result = self.wrap_result(result)

            _pre_wrapped_call = self.prepare_call(call, self.external_di, result)

            args, _call_return = _pre_wrapped_call()

            result = self.reassemble_result(result, args, _call_return)

            result = self.unwrap_result(result)

            self.debug_message("Leaving %s", name)
            return result

        _wrapped.__name__ = name
        _wrapped.__qualname__ = _wrapped.__name__

        return _wrapped


class PipelineExecutionContext(object):
    def add_stage(self, step, pipeline=None):
        if pipeline is None:
            pipeline = Pipeline()

        self.steps[step] = pipeline
        return pipeline

    def get_step_keys(self):
        return list(self.steps.keys())

    def dispatch(self, step, **kwargs):
        return self.pipeline_environment.run(self.steps[step], **kwargs)

    def __new__(cls, *args, **kwargs):
        instance = super(PipelineExecutionContext, cls).__new__(cls)
        instance.steps = {}
        instance.pipeline_environment = PipelineEnvironment()
        return instance



# -*- coding: utf-8 -*-
"""
The graphml module contains output routines to output GraphML structured data from internal graph representations.
"""

import numpy as np

from networkx.readwrite import GraphMLWriter


def to_graphml_writer(g):
    """
    Takes a networkx graph and returns a GraphMLWriter containing the graph.
    
    :param g: graph
    :return: GraphMLWriter instance
    """
    writer = GraphMLWriter()
    writer.xml_type.update({
        np.float32: 'double',
        np.float64: 'double'
    })
    writer.add_graph_element(g)
    return writer


def write_graphml(g, name):
    """
    Writes a networkx graph in GraphML format to a file.
    
    :param g: graph
    :param name: filename
    :return: 
    """
    writer = to_graphml_writer(g)
    writer.dump(name)


def to_graphml_string(g):
    """
    Converts a networkx graph to a GraphML representation.
    
    :param g: graph
    :return: graphml string
    """
    writer = to_graphml_writer(g)
    return str(writer)



# -*- coding: utf-8 -*-
"""
The executor submodule contains the PipelineExecutor, which runs an image processing pipeline.
"""

import gc
import logging
try:
    import tqdm
except ImportError:
    tqdm = None
from itertools import product
from collections import OrderedDict

from copy import deepcopy


from ..misc.processpool import cpu_count, SimpleProcessPool, InProcessFakePool, WrappedException
from .pipeline import PipelineEnvironment


class Collected(object):
    pass


class Every(object):
    pass


class Skip(RuntimeWarning):
    meta = None

    def __init__(self, meta):
        self.meta = meta


class NotDispatchedYet(object):
    pass


exception_debugging = False

singleton_class_mapper_local_cache = {}


def singleton_class_mapper(class_, what, args, kwargs):
    try:
        if class_ not in singleton_class_mapper_local_cache:
            singleton_class_mapper_local_cache[class_] = class_.__new__(class_)

        result = getattr(singleton_class_mapper_local_cache[class_], what)(*args, **kwargs)
        gc.collect()

        return result
    except Exception as _:
        raise


def get_progress_bar(n):
    if tqdm:
        return iter(tqdm.tqdm(range(n)))
    else:
        return iter(range(n))


# noinspection PyMethodMayBeStatic
class PipelineExecutor(object):
    wait = 0.01

    multiprocessing = True

    def __init__(self):
        if self.multiprocessing:
            if self.multiprocessing is True:
                self.multiprocessing = cpu_count()

    def set_workload(self, meta_tuple, item_counts, processing_order):
        self.meta_tuple = meta_tuple
        self.item_counts = item_counts
        self.processing_order = processing_order

    def initialize(self, pec, *args, **kwargs):

        self.log = logging.getLogger(__name__)
        self.pec = pec

        complete_args = (self.pec, '__init__', args, kwargs,)

        self.complete_args = complete_args

        singleton_class_mapper(*complete_args)

        if self.multiprocessing:
            self.pool = SimpleProcessPool(
                processes=self.multiprocessing,
                initializer=singleton_class_mapper,
                initargs=complete_args,
                future_timeout=30.0 * 60,  # five minute timeout, only works with the self-written pool
            )
        else:
            self.pool = InProcessFakePool()

    def set_progress_total(self, l):
        self.progress_indicator = get_progress_bar(l)

    def progress_tick(self):
        try:
            next(self.progress_indicator)
        except StopIteration:
            pass

    # noinspection PyUnusedLocal
    def in_cache(self, token):
        return False

    # noinspection PyUnusedLocal
    def get_cache(self, token):
        pass

    # noinspection PyUnusedLocal
    def set_cache(self, token, result):
        if result is None:
            return

    def skip_callback(self, op, skipped):
        self.log.info("Operation %r caused Skip for %r.", op, skipped)

    def run(self):
        meta_tuple = self.meta_tuple

        if getattr(self, 'cache', False) is False:
            self.cache = False

        sort_order = [index for index, _ in sorted(enumerate(self.processing_order), key=lambda p: p[0])]

        def prepare_steps(step, replace):
            return list(meta_tuple(*t) for t in sorted(product(*[
                self.item_counts[num] if value == replace else [value] for num, value in
                enumerate(step)
                ]), key=lambda t: [t[i] for i in sort_order]))

        todo = OrderedDict()

        reverse_todo = {}
        results = {}

        mapping = {}
        reverse_mapping = {}

        steps = singleton_class_mapper(self.pec, 'get_step_keys', (), {})

        for step in steps:
            order = prepare_steps(step, Every)
            reverse_todo.update({k: step for k in order})
            for k in order:
                todo[k] = NotDispatchedYet
            deps = {t: set(prepare_steps(t, Collected)) for t in order}
            mapping.update(deps)
            for key, value in deps.items():
                for k in value:
                    if k not in reverse_mapping:
                        reverse_mapping[k] = {key}
                    else:
                        reverse_mapping[k] |= {key}

        mapping_copy = deepcopy(mapping)

        def is_concrete(t):
            for n in t:
                if n is Collected or n is Every:
                    return False
            return True

        # initial_length = len(todo)

        self.set_progress_total(len(todo))

        check = OrderedDict()

        cache_originated = set()

        invalidated = set()

        concrete_counter, non_concrete_counter = 0, 0

        while len(todo) > 0 or len(check) > 0:
            for op in list(todo.keys()):

                result = None

                if op not in invalidated:

                    parameter_dict = {'meta': op}

                    if is_concrete(op):
                        # we are talking about a definite point, that is one that is not dependent on others
                        concrete_counter += 1
                        priority = 1 * concrete_counter
                    elif len(mapping[op]) != 0:
                        continue
                    else:
                        collected = OrderedDict()
                        for fetch in sorted(mapping_copy[op], key=lambda t: [t[i] for i in sort_order]):
                            collected[fetch] = results[fetch]
                        parameter_dict[PipelineEnvironment.KEY_COLLECTED] = collected
                        non_concrete_counter += 1
                        priority = -1 * non_concrete_counter

                    token = (reverse_todo[op], op,)
                    if self.in_cache(token):
                        cache_originated.add(op)
                        raise RuntimeError('TODO')  # TODO
                        # result = self.pool.advanced_apply(
                        #     command=singleton_class_mapper,
                        #     args=(self.__class__, 'get_cache', (token,), {},),
                        #     priority=priority
                        # )

                    else:
                        result = self.pool.advanced_apply(
                            singleton_class_mapper,
                            args=(self.pec, 'dispatch', (reverse_todo[op],), parameter_dict,),
                            priority=priority
                        )

                results[op] = result

                check[op] = True

                del todo[op]

            for op in list(check.keys()):
                result = results[op]

                modified = False

                if op in invalidated:
                    if getattr(result, 'fail', False) and callable(result.fail):
                        result.fail()
                        modified = True
                else:

                    if self.wait:
                        result.wait(self.wait)

                    if result.ready():
                        try:
                            result = result.get()

                            if op not in cache_originated:
                                token = (reverse_todo[op], op,)
                                # so far, solely accessing (write) the cache from
                                # one process should mitigate locking issues
                                self.set_cache(token, result)

                        except WrappedException as ee:
                            e = ee.exception
                            if type(e) == Skip:
                                old_invalid = invalidated.copy()

                                def _add_to_invalid(what):
                                    if what not in invalidated:
                                        invalidated.add(what)
                                        if what in mapping_copy:
                                            for item in mapping_copy[what]:
                                                _add_to_invalid(item)

                                _add_to_invalid(e.meta)

                                new_invalid = invalidated - old_invalid

                                self.skip_callback(op, new_invalid)

                            else:
                                if exception_debugging:
                                    raise

                                self.log.exception("Exception occurred at op=%s: %s",
                                                   repr(reverse_todo[op]) + ' ' + repr(op), ee.message)

                            result = None

                        modified = True

                if modified:
                    results[op] = result

                    if op in reverse_mapping:
                        for affected in reverse_mapping[op]:
                            mapping[affected] -= {op}

                    del check[op]
                    self.progress_tick()

        self.progress_tick()

        self.close()

    def close(self):
        self.pool.close()



# -*- coding: utf-8 -*-
"""
The h5writer submodule contains the hdf5_output function, which accepts various datatypes (binary, image, tabular),
and adds them to a (possibly pre-existing) HDF5 file.
"""

from pandas import HDFStore, DataFrame
from tables.nodes import filenode
from tables import NodeError, Filters

from os.path import isfile
from time import sleep
from datetime import datetime
from os import remove, fdopen, open as low_level_open, O_CREAT, O_EXCL, O_WRONLY, getpid
from fnmatch import fnmatch

import numpy as np
import zlib
import bz2
try:
    import lzma
except ImportError:
    lzma = None

import pickle
import time

lock_debugging = False


def wait_for_lock_and_prepare_filename(base_filename, local_timeout):
    found_filename = False
    filename = base_filename
    lock_file = None

    while not found_filename:

        lock_file = '%s.lock' % (filename,)

        begin_time = datetime.now()

        while isfile(lock_file) or local_timeout == 0:
            # print(lock_file, isfile(lock_file), local_timeout)
            sleep(1.0)
            current_time = datetime.now()
            if (current_time - begin_time).total_seconds() > local_timeout:
                newfilename = '%s_%s.h5' % (base_filename, current_time.strftime("%Y%m%d%H%M%S_%f"),)
                print(
                    "WARNING: Waited %d seconds to acquire a lock for \"%s\", "
                    "but failed. Will now try to write data to new file \"%s\"." %
                    (local_timeout, filename, newfilename,)
                )
                filename = newfilename
                lock_file = '%s.lock' % (filename,)
                break

        if not isfile(lock_file):
            found_filename = True
    return filename, lock_file


def acquire_lock(lock_file):
    if lock_debugging:
        print("Process %d acquired lock %s." % (getpid(), lock_file))
    return fdopen(low_level_open(lock_file, O_CREAT | O_EXCL | O_WRONLY), 'w')


def release_lock(lock_file):
    if lock_debugging:
        print("Process %d released lock %s." % (getpid(), lock_file))
    try:
        remove(lock_file)
    except FileNotFoundError:
        pass


timeout = 5 * 60.0


def hdf5_node_name(s):
    s = s.\
        replace('/', '_').\
        replace(':', '_').\
        replace('\\', '_').\
        replace('.', '_').\
        replace('-', '_').\
        replace('?', '_').\
        replace('=', '_').\
        replace('&', '_').\
        replace(';', '_')

    if s[0:1] == '_':
        s = s[1:]

    return s


class CompressedObject(object):
    __slots__ = 'data', 'compression'

    debug = False
    default_compression = 'bz2'

    compressions = {
        'zlib': (lambda _: zlib.compress(_, 6), lambda _: zlib.decompress(_)),
        'bz2': (lambda _: bz2.compress(_, 9), lambda _: bz2.decompress(_)),
        'lzma': (lambda _: lzma.compress(_, check=lzma.CHECK_NONE), lambda _: lzma.decompress(_)),
    }

    def __init__(self, data, compression=None):
        before = time.time()
        if compression is None:
            compression = self.default_compression

        serialized = pickle.dumps(data, pickle.HIGHEST_PROTOCOL)
        before_len = len(serialized)

        assert compression in self.compressions

        compress, uncompress = self.compressions[compression]

        serialized = compress(serialized)

        self.compression = compression
        self.data = serialized

        del serialized
        del data

        after_len = len(self.data)
        after = time.time()

        if self.debug:
            print(
                "... serialized and compressed object using %s from %d to %d (%.3f) took %.3fs" % (
                    compression, before_len, after_len, after_len/before_len, (after-before)
                )
            )

    def get(self):
        compress, uncompress = self.compressions[self.compression]
        return pickle.loads(uncompress(self.data))


def return_or_uncompress(something):
    if isinstance(something, CompressedObject):
        return something.get()
    else:
        return something


def hdf5_output(_filename, immediate_prefix='', tabular_name='result_table'):
    def _inner_hdf5_output(meta, result):

        # noinspection PyProtectedMember
        meta_str = '_'.join(
            k + '_' + ('%09d' % v if type(v) == int else v.__name__)
            for k, v in sorted(meta._asdict().items(), key=lambda x: x[0])
        )

        prefix = '/results/'

        if immediate_prefix != '':
            prefix += immediate_prefix

        prefix += '/' + meta_str + '/'

        success = False

        local_timeout = timeout

        base_filename = _filename

        lock_file = None

        while not success:

            filename, lock_file = wait_for_lock_and_prepare_filename(base_filename, local_timeout)

            compression_type = 'zlib'
            compression_level = 6

            compression_filter = Filters(complib=compression_type, complevel=compression_level)

            try:
                # race conditions
                # open(lock_file, 'w+')
                # noinspection PyUnusedLocal
                with acquire_lock(lock_file) as lock:
                    store = HDFStore(filename, complevel=compression_level, complib=compression_type)

                    # noinspection PyProtectedMember
                    h5 = store._handle

                    # cache for palettes
                    # currently unused
                    # palette_written = {}

                    def store_image(h5path, name, data, upsample_binary=True):
                        h5path = h5path.replace('//', '/')
                        # hdf5 stores bitfields as well, but default 0,1 will be invisible on a fixed 0-255 palette ...
                        if data.dtype == bool and upsample_binary:
                            data = (data * 255).astype(np.uint8)
                        arr = h5.create_carray(h5path, name, obj=data, createparents=True, filters=compression_filter)
                        arr.attrs.CLASS = 'IMAGE'
                        arr.attrs.IMAGE_SUBCLASS = 'IMAGE_GRAYSCALE'
                        arr.attrs.IMAGE_VERSION = '1.2'

                    def store_data(h5path, name, data):
                        h5path = h5path.replace('//', '/')

                        h5path_splits = [x for x in h5path.split('/') if x != '']

                        for i in range(len(h5path_splits)):
                            try:
                                h5.create_group('/' + '/'.join(h5path_splits[:i]), h5path_splits[i])
                            except NodeError:
                                pass

                        f = filenode.new_node(h5, where=h5path, name=name, filters=compression_filter)
                        if type(data) == str:
                            data = data.encode('utf-8')
                        f.write(data)
                        f.close()

                    def store_table(name, data):
                        _frame = DataFrame(data)
                        store[name] = _frame  # .append(name, _frame, data_columns=_frame.columns)

                    image_counter = {}
                    data_counter = {}
                    table_counter = {}

                    def process_row(result_table_rows, m, row):
                        cresults = []

                        # noinspection PyProtectedMember
                        tmp = {('meta_' + mk): (mv if type(mv) == int else -1) for mk, mv in m._asdict().items()}

                        if type(result_table_rows) == list:
                            result_table_rows = {key: True for key in result_table_rows}

                        if '_plain' in result_table_rows:
                            for v in result_table_rows['_plain']:
                                result_table_rows[v] = True
                            del result_table_rows['_plain']

                        def is_wildcarded(s):
                            return '*' in s

                        for k, v in list(result_table_rows.items()):
                            if is_wildcarded(k):
                                del result_table_rows[k]

                                for row_key in row.keys():
                                    if fnmatch(row_key, k):
                                        result_table_rows[row_key] = v

                        for k, v in result_table_rows.items():
                            if v == 'table':
                                if k not in table_counter:
                                    table_counter[k] = 0

                                if k in row and len(row[k]) > 0:
                                    if type(row[k][0]) == list:
                                        # it's a list of lists
                                        # create a mapping table
                                        # point to the mapping table

                                        the_counter = table_counter[k]

                                        new_path = '/tables/_mapping_%s' % (k,)
                                        new_name = '%s_%09d' % (k, the_counter)

                                        tmp[k] = -1
                                        tmp['_mapping_%s' % k] = the_counter

                                        table_counter[k] += 1

                                        i_mapping = []

                                        for n, i_table in enumerate(row[k]):
                                            i_new_path = '/tables/_individual_%s' % (k,)
                                            i_new_name = '%s_%09d' % (k, table_counter[k])
                                            store_table(prefix + i_new_path + '/' + i_new_name, i_table)

                                            i_mapping.append({
                                                '_index': n,
                                                'individual_table': table_counter[k]
                                            })

                                            table_counter[k] += 1

                                        store_table(prefix + new_path + '/' + new_name, i_mapping)

                                        tmp[k] = table_counter[k]
                                        table_counter[k] += 1
                                    else:
                                        new_path = '/tables/%s' % (k,)
                                        new_name = '%s_%09d' % (k, table_counter[k])
                                        store_table(prefix + new_path + '/' + new_name, row[k])
                                        tmp[k] = table_counter[k]
                                        table_counter[k] += 1
                                else:
                                    tmp[k] = table_counter[k]
                                    table_counter[k] += 1
                            elif v == 'image':
                                if k not in image_counter:
                                    image_counter[k] = 0

                                if k in row:
                                    new_path = '/images/%s' % (k,)
                                    new_name = '%s_%09d' % (k, image_counter[k])
                                    store_image(prefix + new_path, new_name, return_or_uncompress(row[k]))
                                tmp[k] = image_counter[k]

                                image_counter[k] += 1
                            elif v == 'data':
                                if k not in data_counter:
                                    data_counter[k] = 0

                                if k in row:
                                    new_path = '/data/%s' % (k,)
                                    new_name = '%s_%09d' % (k, data_counter[k])
                                    store_data(prefix + new_path, new_name, return_or_uncompress(row[k]))
                                tmp[k] = data_counter[k]

                                data_counter[k] += 1
                            else:
                                if k in row:
                                    tmp[k] = row[k]
                                else:
                                    tmp[k] = float('nan')

                        cresults.append(tmp)
                        return cresults

                    if 'collected' in result:
                        collected = []
                        for m, row in result['collected'].items():
                            if tabular_name in row:
                                result_table_rows = row[tabular_name]
                                collected += process_row(result_table_rows, m, row)
                        store_table(prefix + 'result_table_collected', collected)

                    if tabular_name in result:
                        store_table(prefix + 'result_table', process_row(result[tabular_name], meta, result))

                    store.close()

                success = True
            except NodeError:
                print("NodeError Exception occurred while writing, " +
                      "apparently the file has already been used to store similar results.")
                # print("Leaving it LOCKED (remove manually!) and trying to write to another file!")
                local_timeout = 0
                release_lock(lock_file)

            except Exception as e:
                print("Exception occurred while writing results: " + repr(e))
                release_lock(lock_file)
                return
        release_lock(lock_file)

        return result

    return _inner_hdf5_output


# Storing other image data or paletted images

# from tables import Atom
# arr = store._handle.create_carray(h5path, name, Atom.from_dtype(data.dtype), data.shape, createparents=True)
# arr[:] = data[:]

# if data.dtype == bool:
    # arr.attrs.PALETTE = pal.object_id

# if False and True not in palette_written:
#
#     palette = np.zeros((256, 3,), dtype=np.uint8)
#
#     s = 256//3
#
#     for i in range(0, s):
#         palette[i+0*s, :] = [i, 0, 0]
#         palette[i+1*s, :] = [s+i, i, 0]
#         palette[i+2*s, :] = [s+i, s+i, i]
#
#     palette[0, :] = 0
#     palette[1, :] = 255
#
#     # pytables does not seem to support object references
#     # so that nice palette code is actually useless ...
#     pal = store._handle.create_array('/', 'palette', palette, createparents=True)
#     pal.attrs.CLASS = 'PALETTE'
#     pal.attrs.PAL_COLORMODEL = 'RGB'
#     pal.attrs.PAL_TYPE = 'STANDARD8'
#     pal.attrs.PAL_VERSION = '1.2'
#
#     palette_written[True] = True
# #else:
# #    pal = store._handle.get_node('/palette')



# -*- coding: utf-8 -*-
"""
The TIFF reader module implements an ImageStack able to open TIFF image stacks, by using the tifffile module.
"""

from ..imagestack import ImageStack, Dimensions

import warnings

with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from tifffile import TiffFile


class TiffImageStack(ImageStack):
    extensions = ('.tif', '.tiff',)

    priority = 1000

    def open(self, location, **kwargs):
        try:
            self.tiff = TiffFile(location.path, fastij=False)
        except TypeError:
            self.tiff = TiffFile(location.path)

        self.set_dimensions_and_sizes([Dimensions.Time], [len(self.tiff.pages)])

    # noinspection PyProtectedMember
    def notify_fork(self):
        self.tiff._fh.close()
        self.tiff._fh.open()

    def get_data(self, what):
        return self.tiff.pages[what[Dimensions.Time]].asarray()

    def get_meta(self, what):
        try:
            calibration = float(self.parameters['calibration'])
        except KeyError:
            calibration = 1.0

        try:
            interval = float(self.parameters['interval'])
        except KeyError:
            interval = 1.0

        position = self.__class__.Position(x=0.0, y=0.0, z=0.0)
        meta = self.__class__.Metadata(
            time=interval * what[Dimensions.Time],
            position=position,
            calibration=calibration
        )
        return meta




# -*- coding: utf-8 -*-
"""
The helper sub-module contains some auxiliary functions to quickly create and run App objects.
"""

from .application import App


def pilyso_app_generator(**kwargs):
    class _PilysoApp(App):
        def options(self):
            return kwargs
    return _PilysoApp


def pipeline_to_app(pipeline):
    return pilyso_app_generator(name=pipeline.__name__.lower(), description=pipeline.__name__, pipeline=pipeline)


def run_pipeline(pipeline):
    return pipeline_to_app(pipeline)().main()




# -*- coding: utf-8 -*-
"""
The pipeline submodule contains the infrastructure for building image processing pipelines.
"""

from .executor import PipelineExecutor, Every, Collected
from .pipeline import Pipeline, PipelineExecutionContext, PipelineEnvironment



# -*- coding: utf-8 -*-
"""
The regression modules contains some helpers to perform linear fits on data with non-linear begins or ends.
"""

import numpy as np
from scipy.stats import linregress


def find_linear_window(
        x,
        y,
        begin=float('nan'),
        end=float('nan'),
        window=0.1,
        condition=('rvalue', 'gt', 0.95),
        return_begin_end=False,
        return_nan_if_impossible=True
):
    """
    Tries to find a continuous window in x/y which (mostly) follows a linear relation subject to condition.
    
    If window is a float, it is seen as relative length of the input lists.
    Linear regressions will be performed on each window, then the windows will be filtered by the condition
    (eg that they have a rvalue better than 0.95). Then the range between the first and the last window to
    follow these conditions will be used to perform the overall regression.
    
    See also :py:func:`scipy.stats.linregress`
    
    :param x: Input data, independent value
    :param y: Input data, dependent value
    :param begin: 
    :param end: 
    :param window: Window, either 
    :param condition: Condition to check, a tuple of three. The first must be a key of a 
            linear regression result object, the second either 'gt' or 'lt', and the third the value to compare. 
    :param return_begin_end: If true, return the found range as well
    :param return_nan_if_impossible: If True, return NaN if no suitable region was found, otherwise throws RuntimeError
    :return: 
    
    """
    data = np.c_[x, y]

    data[:, 1][~np.isfinite(data[:, 1])] = np.log(np.finfo(data.dtype).eps)

    if type(window) == float:
        window = int(window * data.shape[0])

    if window < 3:
        window = 3

    pairs = [[i, min(i+window, data.shape[0])] for i in range(0, data.shape[0], window)]
    with np.errstate(all='ignore'):
        results = [[b, e, linregress(data[b:e, 0], data[b:e, 1])] for b, e in pairs]

    condition_check = (lambda r: getattr(r, condition[0]) > condition[2]) if condition[1] == 'gt'\
        else (lambda r: getattr(r, condition[0]) < condition[2])

    filtered_results = [[b, e] for b, e, r in results if condition_check(r)]

    if not filtered_results:
        if return_nan_if_impossible:

            regression = results[0][2]

            # noinspection PyProtectedMember,PyProtectedMember
            regression = regression._replace(**{k: float('nan') for k in regression._asdict().keys()})

            if return_begin_end:
                return float('nan'), float('nan'), regression
            else:
                return regression
        else:
            raise RuntimeError('No optimal range for regression found.')

    if np.isnan(begin):
        begin = filtered_results[0][0]

    if np.isnan(end):
        end = filtered_results[-1][1]

    filtered_data = data[begin:end, :]

    with np.errstate(all='ignore'):
        regression = linregress(filtered_data[:, 0], filtered_data[:, 1])

    if return_begin_end:
        return begin, end, regression
    else:
        return regression


def prepare_optimized_regression(x, y):
    """
    First finds an optimal window using :py:func:`find_linear_window`, than performs a linear regression.
    
    :param x: independent variable 
    :param y: dependent variable
    :return: 
    
    >>> x = np.linspace(1, 100, 100)
    >>> y = x * 5 + 10
    >>> y[0:10] = 0  # break our nice linear curve
    >>> prepare_optimized_regression(x, y)
    OrderedDict([('slope', 5.0), ('intercept', 10.0), ('rvalue', 0.99999999999999989), ('pvalue', 0.0), \
('stderr', 7.9423456026468591e-09), ('begin_index', 10), ('end_index', 100), ('begin', 11.0), ('end', 100.0)])
    """
    condition = ('rvalue', 'gt', 0.9)
    begin, end, regression = find_linear_window(x, y, condition=condition, return_begin_end=True)
    # noinspection PyProtectedMember
    regression = regression._asdict()
    regression['begin_index'] = begin
    regression['end_index'] = end

    if end == len(x):
        end -= 1

    if (begin != begin or end != end) or begin < 0 or end >= len(x):
        regression['begin'] = regression['end'] = -1
    else:
        regression['begin'] = x[begin]
        regression['end'] = x[end]
    return regression




# -*- coding: utf-8 -*-
"""
The application module contains classes to quickly bootstrap image processing applications built around pipelines,
as well as some necessary pipeline building blocks.
"""

from .application import App, AppInterface, Meta
# noinspection PyUnresolvedReferences
from ..pipeline import *
from .helper import *

# noinspection PyUnresolvedReferences
from ..pipeline.executor import Skip





# -*- coding: utf-8 -*-
"""
The setup.py file contains meta-information about the package, such as name, description and dependencies.
"""

from setuptools import setup, find_packages
import os
import sys
sys.path.insert(0, '.')

on_rtd = os.environ.get('READTHEDOCS') == 'True'

install_requires = [
    'numpy', 'scipy', 'networkx', 'tables', 'numexpr', 'pandas',
    'tifffile', 'nd2file',
    'mfisp_boxdetection', 'molyso', 'tunable',
    'tqdm'  # nicer progress bars, using the MPLv2+MIT licensed version
] + (['scikit-image>=0.12'] if not on_rtd else [])
# read the docs has a problem installing scikit-image, but doc building will fail if all packages are just mockups ...

import mycelyso

setup(
    name='mycelyso',
    version=mycelyso.__version__,
    description='MYCElium anaLYsis SOftware',
    long_description='see https://github.com/modsim/mycelyso',
    author=mycelyso.__author__,
    author_email='c.sachs@fz-juelich.de',
    url='https://github.com/modsim/mycelyso',
    packages=find_packages(),
    install_requires=install_requires,
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Image Recognition',
    ]
)
